import warnings
warnings.simplefilter('ignore')
import numpy as np
import os
import matplotlib.pyplot as plt
import tensorflow as tf
print(f"tensorflow version: {tf.__version__}")
from tensorflow.keras import layers
from tensorflow.keras import models
from tensorflow.keras.models import Sequential
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Input,Conv2DTranspose,BatchNormalization,ReLU,Conv2D,LeakyReLU
import time

import keras

from IPython import display
import skillsnetwork
print(f"skillsnetwork version: {skillsnetwork.__version__}")

import matplotlib.pyplot as plt
#%matplotlib inline

import os
from os import listdir
from pathlib import Path
import imghdr

from tqdm import tqdm


# This function will allow us to easily plot data taking in x values, y values, and a title
def plot_distribution(real_data, generated_data, discriminator=None, density=True):
    plt.hist(real_data.numpy(), 100, density=density, facecolor='g', alpha=0.75, label='real data')
    plt.hist(generated_data.numpy(), 100, density=density, facecolor='r', alpha=0.75, label='generated data q(z) ')

    if discriminator:
        max_ = np.max([int(real_data.numpy().max()), int(generated_data.numpy().max())])
        min_ = np.min([int(real_data.numpy().min()), int(generated_data.numpy().min())])
        x = np.linspace(min_, max_, 1000).reshape(-1, 1)
        plt.plot(x, tf.math.sigmoid(discriminator(x, training=False).numpy()), label='discriminator', color='k')
        plt.plot(x, 0.5 * np.ones(x.shape), label='0.5', color='b')
        plt.xlabel('x')

    plt.legend()
    plt.show()


def plot_array(X, title=""):
    plt.rcParams['figure.figsize'] = (20, 20)

    for i, x in enumerate(X[0:5]):
        x = x.numpy()
        max_ = x.max()
        min_ = x.min()
        xnew = np.uint(255 * (x - min_) / (max_ - min_))
        plt.subplot(1, 5, i + 1)
        plt.imshow(xnew)
        plt.axis("off")

    plt.show()

mean = [10]
cov = [[1]]
X = tf.random.normal((5000,1),mean=10,stddev=1.0)

print("mean:",np.mean(X))
print("standard deviation:",np.std(X))

Z = tf.random.normal((5000,1),mean=0,stddev=2)

print("mean:",np.mean(Z))
print("standard deviation:",np.std(Z))

plot_distribution(X,Z,discriminator=None,density=True)

Xhat=Z+10

print("mean:",np.mean(Xhat))
print("standard deviation:",np.std(Xhat))

plot_distribution(X,Xhat,discriminator=None,density=True)

def make_generator_model():
    generator = tf.keras.Sequential()
    generator.add(layers.Dense(1))
    return generator

generator=make_generator_model()

Xhat = generator(Z, training=False)
plot_distribution(real_data=X,generated_data=Xhat)

def make_discriminator_model():
    discriminator=tf.keras.Sequential()
    discriminator.add(layers.Dense(1))
    return discriminator

discriminator=make_discriminator_model()

plot_distribution(real_data=X,generated_data=Xhat,discriminator=discriminator)

py_x=tf.math.sigmoid(discriminator(X,training=False))
np.sum(py_x>0.5)

py_x=discriminator(Xhat)
np.sum(py_x>0.5)

def get_accuracy(X,Xhat):
    total=0
    py_x=tf.math.sigmoid(discriminator(X,training=False))
    total=np.mean(py_x)
    py_x=tf.math.sigmoid(discriminator(Xhat,training=False))
    total+=np.mean(py_x)
    return total/2

get_accuracy(X,Xhat)

# This method returns a helper function to compute crossentropy loss
cross_entropy = tf.keras.losses.BinaryCrossentropy(from_logits=True)
def generator_loss(Xhat):
    return cross_entropy(tf.ones_like(Xhat), Xhat)

def discriminator_loss(X, Xhat):
    real_loss = cross_entropy(tf.ones_like(X), X)
    fake_loss = cross_entropy(tf.zeros_like(Xhat), Xhat)
    total_loss = 0.5*(real_loss + fake_loss)
    return total_loss

generator_optimizer = tf.keras.optimizers.Adam(5e-1,beta_1=0.5,beta_2=0.8)

discriminator_optimizer = tf.keras.optimizers.Adam(5e-1,beta_1=0.5, beta_2=0.8)

# parameters for training
epochs = 20
BATCH_SIZE = 5000
noise_dim = 1
epsilon = 100

# discrimator and gernerator
tf.random.set_seed(0)
discriminator = make_discriminator_model()
generator = make_generator_model()

tf.config.run_functions_eagerly(True)

gen_loss_epoch = []
disc_loss_epoch = []
plot_distribution(real_data=X, generated_data=Xhat, discriminator=discriminator)
print("epoch", 0)

for epoch in tqdm(range(epochs)):
    # data for the true distribution of your real data samples training ste
    x = tf.random.normal((BATCH_SIZE, 1), mean=10, stddev=1.0)
    # random samples it was found if you increase the standard deviation, you get better results
    z = tf.random.normal([BATCH_SIZE, noise_dim], mean=0, stddev=10)
    # needed to compute the gradients for a list of variables.
    with tf.GradientTape() as gen_tape, tf.GradientTape() as disc_tape:
        # generated sample
        xhat = generator(z, training=True)
        # the output of the discriminator for real data
        real_output = discriminator(x, training=True)
        # the output of the discriminator  data
        fake_output = discriminator(xhat, training=True)
        # loss for each
        gen_loss = generator_loss(fake_output)
        disc_loss = discriminator_loss(real_output, fake_output)
    # Compute the gradients for gen_loss and generator
    gradients_of_generator = gen_tape.gradient(gen_loss, generator.trainable_variables)
    # Compute the gradients for gen_loss and discriminator
    gradients_of_discriminator = disc_tape.gradient(disc_loss, discriminator.trainable_variables)
    # Ask the optimizer to apply the processed gradients
    generator_optimizer.apply_gradients(zip(gradients_of_generator, generator.trainable_variables))
    discriminator_optimizer.apply_gradients(zip(gradients_of_discriminator, discriminator.trainable_variables))

    # Save and display the generator and discriminator if the performance increases
    if abs(0.5 - get_accuracy(x, xhat)) < epsilon:
        epsilon = abs(0.5 - get_accuracy(x, xhat))
        generator.save('generator')
        discriminator.save('discriminator')
        print(get_accuracy(x, xhat))
        plot_distribution(real_data=X, generated_data=xhat, discriminator=discriminator)
        print("epoch", epoch)

        generator = make_generator_model()
        generator = models.load_model('generator')
        xhat = generator(z)
        discriminator = models.load_model('discriminator')
        plot_distribution(real_data=X, generated_data=xhat, discriminator=discriminator)

        dataset_url = "https://cf-courses-data.s3.us.cloud-object-storage.appdomain.cloud/IBMDeveloperSkillsNetwork-ML311-Coursera/labs/Module6/cartoon_20000.zip"
        await skillsnetwork.prepare(dataset_url, overwrite=True)

        img_height, img_width, batch_size = 70, 70, 150

        train_ds = tf.keras.utils.image_dataset_from_directory(directory='cartoon_20000',
                                                               image_size=(img_height, img_width),
                                                               batch_size=batch_size,
                                                               label_mode=None)

        normalization_layer = layers.experimental.preprocessing.Rescaling(scale=1. / 127.5, offset=-1)
        normalized_ds = train_ds.map(lambda x: normalization_layer(x))

        images = train_ds.take(1)

        X = [x for x in images]

        plot_array(X[0])


        def make_generator():
            model = Sequential()

            # input is latent vector of 100 dimensions
            model.add(Input(shape=(1, 1, 100), name='input_layer'))

            # Block 1 dimensionality of the output space  64 * 8
            model.add(Conv2DTranspose(64 * 8, kernel_size=4, strides=4, padding='same',
                                      kernel_initializer=tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02),
                                      use_bias=False, name='conv_transpose_1'))
            model.add(BatchNormalization(momentum=0.1, epsilon=0.8, center=1.0, scale=0.02, name='bn_1'))
            model.add(ReLU(name='relu_1'))

            # Block 2: input is 4 x 4 x (64 * 8)
            model.add(Conv2DTranspose(64 * 4, kernel_size=4, strides=2, padding='same',
                                      kernel_initializer=tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02),
                                      use_bias=False, name='conv_transpose_2'))
            model.add(BatchNormalization(momentum=0.1, epsilon=0.8, center=1.0, scale=0.02, name='bn_2'))
            model.add(ReLU(name='relu_2'))

            # Block 3: input is 8 x 8 x (64 * 4)
            model.add(Conv2DTranspose(64 * 2, kernel_size=4, strides=2, padding='same',
                                      kernel_initializer=tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02),
                                      use_bias=False, name='conv_transpose_3'))
            model.add(BatchNormalization(momentum=0.1, epsilon=0.8, center=1.0, scale=0.02, name='bn_3'))
            model.add(ReLU(name='relu_3'))

            # Block 4: input is 16 x 16 x (64 * 2)
            model.add(Conv2DTranspose(64 * 1, kernel_size=4, strides=2, padding='same',
                                      kernel_initializer=tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02),
                                      use_bias=False, name='conv_transpose_4'))
            model.add(BatchNormalization(momentum=0.1, epsilon=0.8, center=1.0, scale=0.02, name='bn_4'))
            model.add(ReLU(name='relu_4'))

            model.add(Conv2DTranspose(3, 4, 2, padding='same',
                                      kernel_initializer=tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02),
                                      use_bias=False,
                                      activation='tanh', name='conv_transpose_5'))

            return model


        gen = make_generator()
        gen.summary()


        def make_discriminator():
            model = Sequential()

            # Block 1: input is 64 x 64 x (3)
            model.add(Input(shape=(64, 64, 3), name='input_layer'))
            model.add(Conv2D(64, kernel_size=4, strides=2, padding='same',
                             kernel_initializer=tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02),
                             use_bias=False, name='conv_1'))
            model.add(LeakyReLU(0.2, name='leaky_relu_1'))

            # Block 2: input is 32 x 32 x (64)
            model.add(Conv2D(64 * 2, kernel_size=4, strides=2, padding='same',
                             kernel_initializer=tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02),
                             use_bias=False, name='conv_2'))
            model.add(BatchNormalization(momentum=0.1, epsilon=0.8, center=1.0, scale=0.02, name='bn_1'))
            model.add(LeakyReLU(0.2, name='leaky_relu_2'))

            # Block 3
            model.add(Conv2D(64 * 4, 4, 2, padding='same',
                             kernel_initializer=tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02),
                             use_bias=False, name='conv_3'))
            model.add(BatchNormalization(momentum=0.1, epsilon=0.8, center=1.0, scale=0.02, name='bn_2'))
            model.add(LeakyReLU(0.2, name='leaky_relu_3'))

            # Block 4
            model.add(Conv2D(64 * 8, 4, 2, padding='same',
                             kernel_initializer=tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02),
                             use_bias=False, name='conv_4'))
            model.add(BatchNormalization(momentum=0.1, epsilon=0.8, center=1.0, scale=0.02, name='bn_3'))
            model.add(LeakyReLU(0.2, name='leaky_relu_4'))

            # Block 5
            model.add(Conv2D(1, 4, 2, padding='same',
                             kernel_initializer=tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02),
                             use_bias=False,
                             activation='sigmoid', name='conv_5'))

            return model


        disc = make_discriminator()
        disc.summary()

        cross_entropy = tf.keras.losses.BinaryCrossentropy(from_logits=True)


        def generator_loss(Xhat):
            return cross_entropy(tf.ones_like(Xhat), Xhat)


        def discriminator_loss(X, Xhat):
            real_loss = cross_entropy(tf.ones_like(X), X)
            fake_loss = cross_entropy(tf.zeros_like(Xhat), Xhat)
            total_loss = 0.5 * (real_loss + fake_loss)
            return total_loss


        learning_rate = 0.0002

        generator_optimizer = tf.keras.optimizers.Adam(lr=0.0002, beta_1=0.5, beta_2=0.999)

        discriminator_optimizer = tf.keras.optimizers.Adam(lr=0.0002, beta_1=0.5, beta_2=0.999)


        @tf.function
        def train_step(X):
            # random samples it was found if you increase the  stander deviation, you get better results
            z = tf.random.normal([BATCH_SIZE, 1, 1, latent_dim])
            # needed to compute the gradients for a list of variables.
            with tf.GradientTape() as gen_tape, tf.GradientTape() as disc_tape:
                # generated sample
                xhat = generator(z, training=True)
                # the output of the discriminator for real data
                real_output = discriminator(X, training=True)
                # the output of the discriminator for fake data
                fake_output = discriminator(xhat, training=True)

                # loss for each
                gen_loss = generator_loss(fake_output)
                disc_loss = discriminator_loss(real_output, fake_output)
            # Compute the gradients for gen_loss and generator

            gradients_of_generator = gen_tape.gradient(gen_loss, generator.trainable_variables)
            # Compute the gradients for gen_loss and discriminator
            gradients_of_discriminator = disc_tape.gradient(disc_loss, discriminator.trainable_variables)
            # Ask the optimizer to apply the processed gradients
            generator_optimizer.apply_gradients(zip(gradients_of_generator, generator.trainable_variables))
            discriminator_optimizer.apply_gradients(zip(gradients_of_discriminator, discriminator.trainable_variables))

            generator = make_generator()
            BATCH_SIZE = 128

            latent_dim = 100
            noise = tf.random.normal([BATCH_SIZE, 1, 1, latent_dim])
            Xhat = generator(noise, training=False)
            plot_array(Xhat)

            epochs = 1

            discriminator = make_discriminator()

            generator = make_generator()

            for epoch in range(epochs):

                # data for the true distribution of your real data samples training ste
                start = time.time()
                i = 0
                for X in tqdm(normalized_ds, desc=f"epoch {epoch + 1}", total=len(normalized_ds)):

                    i += 1
                    if i % 1000:
                        print("epoch {}, iteration {}".format(epoch + 1, i))

                    train_step(X)

                noise = tf.random.normal([BATCH_SIZE, 1, 1, latent_dim])
                Xhat = generator(noise, training=False)
                X = [x for x in normalized_ds]
                print("orignal images")
                plot_array(X[0])
                print("generated images")
                plot_array(Xhat)
                print('Time for epoch {} is {} sec'.format(epoch + 1, time.time() - start))


generator_url="https://cf-courses-data.s3.us.cloud-object-storage.appdomain.cloud/IBM-GPXX0XCEEN/data/generator.tar"
await skillsnetwork.prepare(generator_url, overwrite=True)

from tensorflow.keras.models import load_model


full_generator=load_model("generator")

latent_dim=100

# input consists of noise vectors
noise = tf.random.normal([200, 1, 1, latent_dim])

# feed the noise vectors to the generator
Xhat=full_generator(noise,training=False)
plot_array(Xhat)

for c in [1,0.8,0.6,0.4]:
    Xhat=full_generator(c*tf.ones([1, 1, 1, latent_dim]),training=False) # latent_dim = 100 defined previously
    plot_array(Xhat)


z=np.ones( (1, 1, 1, latent_dim))
for n in range(10):

    z[0, 0, 0, 0:10*n]=-1

    Xhat=full_generator(z,training=False)
    print("elements from 0 to {} is set to -1".format(10*n))
    plot_array(Xhat)

    for n in range(10):
        z = np.random.normal(0, 1, (1, 1, 1, latent_dim))

        z[0, 0, 0, 0:35] = 1

        Xhat = full_generator(z, training=False)

        plot_array(Xhat)




import tokenizer
from torch import tensor
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
model_name = "facebook/blenderbot-400M-distill"
# Load model (download on first run and reference local installation for consequent runs)
model = AutoModelForSeq2SeqLM.from_pretrained(model_name)
tokenizer = AutoTokenizer.from_pretrained(model_name)
conversation_history = []
history_string = "\n".join(conversation_history)
history_string
input_text ="hello"
input_text
inputs = tokenizer.encode_plus(history_string, input_text, return_tensors="pt")
inputs
tokenizer.pretrained_vocab_files_map
outputs = model.generate(**inputs)
outputs
response = tokenizer.decode(outputs[0], skip_special_tokens=True).strip()
response
conversation_history.append(input_text)
conversation_history.append(response)
conversation_history
while True:
    # Create conversation history string
    history_string = "\n".join(conversation_history)

    # Get the input data from the user
    input_text = input("> ")

    # Tokenize the input text and history
    inputs = tokenizer.encode_plus(history_string, input_text, return_tensors="pt")

    # Generate the response from the model
    outputs = model.generate(**inputs)

    # Decode the response
    response = tokenizer.decode(outputs[0], skip_special_tokens=True).strip()

    # Add interaction to conversation history
    conversation_history.append(input_text)
    conversation_history.append(response)



import json
import copy
import ast
import pandas as pd
from urllib3.util import Retry
from robot.api import logger
from cqepyutils.PandasUtils import read_csv, filter_df, reset_index
from RequestsLibrary import RequestsLibrary  # Add this import statement


def trigger_api(test_data_full_path, test_input_path, api_name, base_url, auth=None):
    logger.info('*************************************************************************************************')
    logger.info(f'Trigger API using the parameters specified in the test data sheet - {api_name}')
    logger.info('*************************************************************************************************')

    logger.info('Step-01: Start - Read and filter test data sheet based on API_NAME')
    # Example of using functions from the imported module directly
    test_cases_df = read_csv(str(test_data_full_path), dtype=str)
    selected_api_df = filter_df(test_cases_df, 'test_name', api_name).copy()
    selected_api_df = filter_df(selected_api_df, 'tbe', 'YES')
    selected_api_df = reset_index(selected_api_df)
    logger.debug(f'API Name      : {api_name}')
    logger.info('Step-01: End   - Read and filter test data sheet based on API_NAME')

    logger.info('Step-02: Start - Identify the request type')
    request_type = str(selected_api_df['request_type'][0])
    logger.info('Step-02: End   - Identify the request type')

    logger.info('Step-03: Start - Assign default values to None')
    params = None
    data = None
    # auth = None
    files = None
    json_data = None
    cookies = None
    timeout = None
    proxies = None
    verify = False
    logger.info('Step-03: End   - Assign default values to None')

    logger.info('Step-04: Start - Assign project specific values')
    # Check if authorization is required
    if auth is not None:
        auth = auth.split(',')
        logger.debug(auth)
    elif pd.isna(selected_api_df.loc[0, 'auth']):
        auth = None
    else:
        auth = str(selected_api_df['auth'].item()).split(',')
    logger.info('Step-04: End   - Assign project specific values')

    logger.info('Step-05: Start - Identify uri')

    # Unique identifier for creating a session
    alias = str(api_name)

    uri = str(selected_api_df['uri'][0])
    logger.info('Step-05: End   - Identify uri')

    logger.info('Step-06: Start - Update the path parameter in URI')
    # Identify the path parameters specific columns and create a separate data frame
    selected_api_flt_pp_df = selected_api_df.filter(regex='_pp$', axis=1)

    # Identify the path parameters specific for the api
    selected_api_flt_pp_df = selected_api_flt_pp_df.dropna(axis='columns')

    # Path Params value which needs to be passed should be as dictionary
    if selected_api_flt_pp_df.empty:
        path_params = None
    else:
        path_params_list = selected_api_flt_pp_df.to_dict('records')[0]

        # Remove _pp from the parameter name
        path_params = {x.replace('_pp', ''): v for x, v in path_params_list.items()}

        # Iterate through each values and update the URL
        for key, value in path_params.items():
            uri = uri.replace(str('{') + key + str('}'), value)
        logger.debug(uri)

    logger.info('Step-06: End   - Update the path parameter in URI')

    logger.info('Step-07: Start - Create the parameter and assign to params')

    # Identify the parameters specific columns and create a separate data frame
    selected_api_flt_df = selected_api_df.filter(regex='_d$', axis=1)

    # Identify the parameters specific for the api
    selected_api_flt_df = selected_api_flt_df.dropna(axis='columns')

    # Params value which needs to be passed should be as dictionary
    if selected_api_flt_df.empty:
        params = None
    else:
        params_list = selected_api_flt_df.to_dict('records')[0]

        # Remove _d from the parameter name
        params = {x.replace('_d', ''): v for x, v in params_list.items()}
    logger.info('Step-07: End   - Create the parameter and assign to params')

    logger.info('Step-08: Start - Extract other additional information required - data, auth, headers, files')

    # Check if data level json is required (body)
    if pd.isna(selected_api_df.loc[0, 'data']):
        data = None
    else:
        file_name = str(selected_api_df['data'][0])
        with open(test_input_path + api_name + '/' + file_name + '.JSON') as f:
            data = json.load(f)

    # Check if headers is required
    headers = {}
    if pd.isna(selected_api_df.loc[0, 'headers']):
        headers = {}
    else:
        headers = {'Content-Type': 'values', 'Accept': 'values'}
        headers_type = str(selected_api_df['headers'].item())
        headers['Content-Type'] = headers_type
        headers['Accept'] = headers_type

    # Check if files is required
    files_str = ''
    if pd.isna(selected_api_df.loc[0, 'files']):
        files = None
    else:
        # Extract the files column content
        files_str = str(selected_api_df['files'][0])

        # Convert the variable to dict
        files = ast.literal_eval(files_str)

        # Iterate through the dict and open the files specified in the data sheet
        # Files will be opened in binary mode and passed to post
        for k, value in files.items():
            f = open(test_input_path + api_name + '/' + value, 'rb')
            files[k] = f

    # List of parameters required for any type of API
    logger.info('type      :' + request_type)
    logger.info('url       :' + base_url)
    logger.info('uri       :' + uri)
    logger.info('params    :' + str(params))
    logger.info('data      :' + str(data))
    logger.info('auth      :' + str(auth))
    logger.info('headers   :' + str(headers))
    logger.info('files     :' + str(files_str))

    # Static variables for the request
    max_retries = 3
    backoff_factor = 0.1
    disable_warnings = 0
    debug = 0

    logger.info('Step-08: End   - Extract other additional information required - data, auth, headers, files')

    logger.info('Step-09: Start - Trigger the API, extract the response and status code')

    # Condition to check GET/POST method
    if request_type == 'GET':

        # Response with status code, response content and api name will be returned as data frame
        # Create Session
        super(RequestsLibrary, self).create_session(alias=alias, url=base_url, headers=headers, cookies=cookies,
                                                    auth=auth, timeout=timeout, proxies=proxies, verify=verify,
                                                    debug=debug, max_retries=max_retries,
                                                    backoff_factor=backoff_factor,
                                                    disable_warnings=1)
        # Trigger GET method
        response = super(RequestsLibrary, self).get_request(alias=alias, uri=uri, headers=headers, json=data,
                                                            params=params, allow_redirects=None, timeout=timeout)
    elif request_type == 'PUT':

        # Response with status code, response content and api name will be returned as data frame
        # Create Session
        super(RequestsLibrary, self).create_session(alias=alias, url=base_url, headers=headers, cookies=cookies,
                                                    auth=auth, timeout=timeout, proxies=proxies, verify=verify,
                                                    debug=debug, max_retries=max_retries,
                                                    backoff_factor=backoff_factor,
                                                    disable_warnings=1)
        # Trigger PUT method
        response = super(RequestsLibrary, self).put_request(alias=alias, uri=uri, data=data, json=json_data,
                                                            params=params, files=files, headers=headers,
                                                            allow_redirects=None, timeout=timeout)
    elif request_type == 'POST':
        # Create Session
        super(RequestsLibrary, self).create_session(alias=alias, url=base_url, headers=headers, cookies=cookies,
                                                    auth=auth, timeout=timeout, proxies=proxies, verify=verify,
                                                    debug=debug, max_retries=max_retries,
                                                    backoff_factor=backoff_factor,
                                                    disable_warnings=1)
        # Trigger POST method
        response = super(RequestsLibrary, self).post_request(alias=alias, uri=uri, data=data, json=json_data,
                                                             params=params, headers=headers, files=files,
                                                             allow_redirects=None, timeout=timeout)
    elif request_type == 'DELETE':
        # Create Session
        super(RequestsLibrary, self).create_session(alias=alias, url=base_url, headers=headers, cookies=cookies,
                                                    auth=auth, timeout=timeout, proxies=proxies, verify=verify,
                                                    debug=debug, max_retries=max_retries,
                                                    backoff_factor=backoff_factor,
                                                    disable_warnings=1)
        # Trigger DELETE method
        response = super(RequestsLibrary, self).delete_request(alias=alias, uri=uri, data=data, json=json_data,
                                                               params=params, headers=headers, allow_redirects=None,
                                                               timeout=timeout)

    selected_api_df.loc[0, 'API_Response_Code'] = response.status_code
    if response.text != '':
        response_string = str(response.content.decode("utf-8"))
        selected_api_df.loc[0, 'API_Response_Content'] = response_string
    else:
        selected_api_df.loc[0, 'API_Response_Content'] = ''
    logger.info('Step-09: End   - Trigger the API, extract the response and status code')
    logger.info('*************************************************************************************************')

    return selected_api_df


if __name__ == "__main__":
    # If you have any code outside the function, keep it here
    pass



import requests
import json

# Zephyr API endpoint and bearer token
zephyr_base_url = 'https://your-jira-instance/rest/zapi/latest'
bearer_token = 'your_bearer_token'


# Create a new test case
def create_test_case(project_id, summary, description, issue_type_id):
    create_test_case_url = f"{zephyr_base_url}/teststep/{project_id}/"

    payload = {
        "fields": {
            "project": {
                "id": project_id
            },
            "summary": summary,
            "description": description,
            "issuetype": {
                "id": issue_type_id
            }
        }
    }

    headers = {
        'Content-Type': 'application/json',
        'Authorization': f'Bearer {bearer_token}'
    }

    response = requests.post(create_test_case_url, headers=headers, data=json.dumps(payload))

    if response.status_code == 201:
        print("Test case created successfully.")
    else:
        print("Failed to create test case.")
        print(response.text)


# Update an existing test case
def update_test_case(test_case_id, new_summary):
    update_test_case_url = f"{zephyr_base_url}/teststep/{test_case_id}/"

    payload = {
        "fields": {
            "summary": new_summary
        }
    }

    headers = {
        'Content-Type': 'application/json',
        'Authorization': f'Bearer {bearer_token}'
    }

    response = requests.put(update_test_case_url, headers=headers, data=json.dumps(payload))

    if response.status_code == 200:
        print("Test case updated successfully.")
    else:
        print("Failed to update test case.")
        print(response.text)


# Update execution results for a test case
def update_execution_results(execution_id, status):
    update_execution_url = f"{zephyr_base_url}/execution/{execution_id}/execute"

    payload = {
        "status": status
    }

    headers = {
        'Content-Type': 'application/json',
        'Authorization': f'Bearer {bearer_token}'
    }

    response = requests.put(update_execution_url, headers=headers, data=json.dumps(payload))

    if response.status_code == 200:
        print("Execution results updated successfully.")
    else:
        print("Failed to update execution results.")
        print(response.text)


# Get project ID by project name
def get_project_id_by_name(project_name):
    get_project_url = f"{zephyr_base_url}/project"

    headers = {
        'Content-Type': 'application/json',
        'Authorization': f'Bearer {bearer_token}'
    }

    response = requests.get(get_project_url, headers=headers)

    if response.status_code == 200:
        projects = response.json()
        for project in projects:
            if project["name"] == project_name:
                return project["id"]
        print(f"Project with name '{project_name}' not found.")
        return None
    else:
        print("Failed to fetch projects.")
        print(response.text)
        return None


# Get test case ID by test case name and folder name
def get_test_case_id_by_name(test_case_name, folder_name=None):
    get_test_cases_url = f"{zephyr_base_url}/teststep"

    headers = {
        'Content-Type': 'application/json',
        'Authorization': f'Bearer {bearer_token}'
    }

    response = requests.get(get_test_cases_url, headers=headers)

    if response.status_code == 200:
        test_cases = response.json()
        for test_case in test_cases:
            if test_case["fields"]["summary"] == test_case_name:
                if folder_name:
                    if "folder" in test_case["fields"] and test_case["fields"]["folder"]:
                        if folder_name == test_case["fields"]["folder"]["folderName"]:
                            return test_case["id"]
                    continue
                else:
                    return test_case["id"]

        print(f"Test case with name '{test_case_name}' not found.")
        return None
    else:
        print("Failed to fetch test cases.")
        print(response.text)
        return None


if __name__ == "__main__":
    # Replace these values with your own
    project_name = "Your_Project_Name"
    test_case_name = "Your_Test_Case_Name"
    folder_name = "Your_Folder_Name"  # Optional, remove or set to None if not using folders

    # Get project ID by project name
    project_id = get_project_id_by_name(project_name)
    if project_id:
        print(f"Project ID for '{project_name}': {project_id}")

    # Get test case ID by test case name (with optional folder name)
    test_case_id = get_test_case_id_by_name(test_case_name, folder_name)
    if test_case_id:
        print(f"Test Case ID for '{test_case_name}': {test_case_id}")



import duckdb
import pandas as pd
import os, time
import logging
import csv
import dask.dataframe as dd

# Setting up logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

# Use a disk-based database instead of in-memory
DATABASE = 'C:/Users/ranji/OneDrive/PycharmProjects/pythontest/results/my_duckdb_database.db'
# Declare variables
SRC_PATH = 'C:/Users/ranji/OneDrive/PycharmProjects/pythontest/src/5m Sales Records.csv'
TRG_PATH = 'C:/Users/ranji/OneDrive/PycharmProjects/pythontest/trg/5m Sales Records.csv'
TABLE_SRC = 'table_src'
TABLE_TRG = 'table_trg'
PRIMARY_KEYS = ['Order Date']
logging.info('Variables declared')


def load_data_into_duckdb(file_path, table_name):
    """Load data from a file into DuckDB based on its format"""
    # Drop the table if it already exists
    conn.execute(f"DROP TABLE IF EXISTS {table_name}")

    extension = os.path.splitext(file_path)[1].lower()
    if extension == '.csv':
        temp_table_name = table_name + '_temp'
        conn.execute(f"DROP TABLE IF EXISTS {temp_table_name}") # Drop the temporary table if it already exists
        conn.execute(f"CREATE TABLE {temp_table_name} AS SELECT * FROM read_csv_auto('{file_path}')")
        conn.execute(f"CREATE TABLE {table_name} AS SELECT * FROM {temp_table_name}")
        conn.execute(f"DROP TABLE {temp_table_name}") # Drop the temporary table
    else:
        raise ValueError(f'Unsupported file format: {extension}')
    column_info = conn.execute(f'PRAGMA table_info({table_name})').fetchall()
    return [col[1] for col in column_info]


logging.info('Function load_data_into_duckdb defined')

# Create a connection to DuckDB
conn = duckdb.connect(database=DATABASE, read_only=False)
logging.info('Connected to DuckDB')

# Load the data into DuckDB based on file format
columns_src_names = load_data_into_duckdb(SRC_PATH, TABLE_SRC)
columns_trg_names = load_data_into_duckdb(TRG_PATH, TABLE_TRG)
logging.info('Data loaded into DuckDB')

# Define keys string for SQL queries
keys_str = ' AND '.join([f'a."{key}" = b."{key}"' if ' ' in key else f'a.{key} = b.{key}' for key in PRIMARY_KEYS])

# Dictionary to store column comparison results
column_differences = {}

# Compare columns using hashes
start_time = time.time()
for col in columns_src_names:
    col_quoted = f'"{col}"' if ' ' in col else col
    src_data = conn.execute(f"SELECT {col_quoted} FROM {TABLE_SRC} ORDER BY {col_quoted}").fetchall()
    trg_data = conn.execute(f"SELECT {col_quoted} FROM {TABLE_TRG} ORDER BY {col_quoted}").fetchall()

    src_hash = hash(str(src_data))
    trg_hash = hash(str(trg_data))

    if src_hash != trg_hash:
        mismatched_rows = conn.execute(
            f"SELECT COUNT(*) FROM {TABLE_SRC} AS a JOIN {TABLE_TRG} AS b ON ({keys_str}) WHERE a.{col_quoted} != b.{col_quoted}").fetchone()[
            0]
        total_rows = conn.execute(f"SELECT COUNT(*) FROM {TABLE_SRC}").fetchone()[0]
        percentage_difference = (mismatched_rows / total_rows) * 100
        column_differences[col] = percentage_difference

# Output the column differences to CSV
with open('total_data_breaks.csv', 'w', newline='') as csvfile:
    writer = csv.writer(csvfile)
    writer.writerow(['Column Name', 'Percentage Difference'])
    for col, diff in column_differences.items():
        writer.writerow([col, f"{diff:.2f}%"])

end_time = time.time()
time_taken = end_time - start_time # Calculate the time taken
logging.info(f'"Time Taken" comparison completed in {time_taken:.2f} seconds')
logging.info('Data breaks report saved as total_data_breaks.csv')

# Only in Source
start_time = time.time()

only_in_source_query = f"""
SELECT a.*
FROM {TABLE_SRC} AS a
LEFT JOIN {TABLE_TRG} AS b ON ({keys_str})
WHERE b."{PRIMARY_KEYS[0]}" IS NULL
"""

only_in_source = conn.execute(only_in_source_query).fetchall()
dd.from_pandas(pd.DataFrame(only_in_source, columns=columns_src_names), npartitions=10).compute().to_csv(
    "only_in_source.csv", index=False)
end_time = time.time()
time_taken = end_time - start_time # Calculate the time taken
logging.info(f'"Time Taken" comparison completed in {time_taken:.2f} seconds')
logging.info('Only in Source.csv')

# Only in Target
start_time = time.time()

only_in_target_query = f"""
SELECT b.*
FROM {TABLE_TRG} AS b
LEFT JOIN {TABLE_SRC} AS a ON ({keys_str})
WHERE a."{PRIMARY_KEYS[0]}" IS NULL
"""

only_in_target = conn.execute(only_in_target_query).fetchall()
dd.from_pandas(pd.DataFrame(only_in_target, columns=columns_trg_names), npartitions=10).compute().to_csv(
    "only_in_target.csv", index=False)

end_time = time.time()
time_taken = end_time - start_time # Calculate the time taken
logging.info(f'"Time Taken" comparison completed in {time_taken:.2f} seconds')
logging.info('Only in Target.csv')

# Source Key Duplicate
start_time = time.time()

source_key_duplicate_query = f"""
SELECT a.*
FROM {TABLE_SRC} AS a
JOIN (
    SELECT "{PRIMARY_KEYS[0]}"
    FROM {TABLE_SRC}
    GROUP BY "{PRIMARY_KEYS[0]}"
    HAVING COUNT(*) > 1
) AS b ON {' AND '.join([f'a."{key}" = b."{key}"' for key in PRIMARY_KEYS])}
"""

source_key_duplicate = conn.execute(source_key_duplicate_query).fetchall()
dd.from_pandas(pd.DataFrame(source_key_duplicate, columns=columns_src_names), npartitions=10).compute().to_csv(
    'source_key_duplicate.csv', index=False)

end_time = time.time()
time_taken = end_time - start_time # Calculate the time taken
logging.info(f'"Time Taken" comparison completed in {time_taken:.2f} seconds')
logging.info('Duplicate in Source.csv')

# Target Key Duplicate
start_time = time.time()

target_key_duplicate_query = f"""
SELECT a.*
FROM {TABLE_TRG} AS a
JOIN (
    SELECT "{PRIMARY_KEYS[0]}"
    FROM {TABLE_TRG}
    GROUP BY "{PRIMARY_KEYS[0]}"
    HAVING COUNT(*) > 1
) AS b ON {' AND '.join([f'a."{key}" = b."{key}"' for key in PRIMARY_KEYS])}
"""

target_key_duplicate = conn.execute(target_key_duplicate_query).fetchall()
dd.from_pandas(pd.DataFrame(target_key_duplicate, columns=columns_trg_names), npartitions=10).compute().to_csv(
    'target_key_duplicate.csv', index=False)

end_time = time.time()
time_taken = end_time - start_time # Calculate the time taken
logging.info(f'"Time Taken" comparison completed in {time_taken:.2f} seconds')
logging.info('Duplicate in Target.csv')

logging.info('Computed data for "Only in Source," "Only in Target," "Source Key Duplicate," and "Target Key Duplicate"')

# Compute metrics and write to separate CSVs
summary_data = {
    'Source Total': conn.execute(f"SELECT COUNT(*) FROM {TABLE_SRC}").fetchone()[0],
    'Target Total': conn.execute(f"SELECT COUNT(*) FROM {TABLE_TRG}").fetchone()[0],
    'Total Data Breaks (Columns)': len(column_differences),
    'Only in Source': len(only_in_source),
    'Only in Target': len(only_in_target),
    'Source Key Duplicate': len(source_key_duplicate),
    'Target Key Duplicate': len(target_key_duplicate)
}

# Convert the summary data to a DataFrame and save as CSV
pd.DataFrame([summary_data]).to_csv('summary.csv', index=False)
logging.info('Summary data computed and saved')

# Close the DuckDB connection
conn.close()
logging.info('DuckDB connection closed')



from robot.api import ExecutionResult, ResultVisitor
import csv
import pandas as pd


class RFUtils:

    def __init__(self):
        pass

    def email_trigger(self, xml_file, csv_file, html_file):
        class TestMetrics(ResultVisitor):

            def visit_test(self, test):
                wr.writerow([test.name, test.doc, test.status, test.message, test.elapsedtime / float(1000)])

        class SuiteMetrics(ResultVisitor):

            def visit_suite(self, suite):
                stats = result.suite.statistics
                wr1.writerow(
                    [suite.name, suite.status, stats.total, stats.passed, stats.failed,
                     suite.elapsedtime / float(60000)])

        op_f = open(csv_file, 'w')
        wr = csv.writer(op_f)

        wr.writerow(['TESTCASE_NAME', 'DOCUMENTATION', 'TESTCASE_STATUS', 'FAILURE_ANALYSIS', 'TESTCASE_ELAPSEDTIME'])
        result = ExecutionResult(xml_file)
        print('after result')
        result.visit(TestMetrics())
        op_f.close()

        suite_csv = csv_file.replace('.csv', '_suite.csv')
        op_f1 = open(suite_csv, 'w')
        wr1 = csv.writer(op_f1)
        wr1.writerow(['SUITE_NAME', 'SUITE_STATUS', 'TOTAL_TESTCASES', 'TESTCASES_PASSED', 'TESTCASES_FAILED',
                      'EXECUTION_TIME'])

        result.visit(SuiteMetrics())
        op_f1.close()
        self.csv_to_html(csv_file, suite_csv, html_file, 'TESTCASE_STATUS')

    def csv_to_html(self, csv_file, suite_csv, html_file, highlight_column_name):
        df = pd.read_csv(csv_file, index_col=False)
        df = df[~df.TESTCASE_NAME.str.contains("Placeholder Test")]
        df.fillna('', inplace=True)
        print(df)

        df1 = pd.read_csv(suite_csv, index_col=False)
        print(df1)

        html = df.style.applymap(self.highlight_vals, subset=[highlight_column_name]). \
            applymap(self.center_vals, subset=['TESTCASE_STATUS']). \
            applymap(self.center_vals, subset=['TESTCASE_ELAPSEDTIME']). \
            applymap(self.left_vals, subset=['DOCUMENTATION']).set_table_styles(
            [{'selector': 'tr.hover td', 'props': [('background-color', 'lightyellow')]},
             {'selector': 'th, td', 'props': [('border', '1px solid black'),
                                              ('padding', '14px'), ('text-align', 'left')]},
             {'selector': 'th', 'props': [('font-family', 'Century Gothic'), ('font-size', '9pt')]},
             {'selector': 'thead', 'props': [('background-color', 'lightblue')]},
             {'selector': '', 'props': [('border-collapse', 'collapse'),
                                        ('border', '1px solid black')]},
             ]).set_properties(**{'font-size': '10pt', 'font-family': 'Century Gothic'}).render()
        html = html.replace('<th class="blank level0 ></th>', '<th class="blank level0">S_NO</th>')

        html1 = df1.style.applymap(self.highlight_cols_red, subset=['TESTCASES_FAILED']). \
            applymap(self.center_vals, subset=['SUITE_STATUS']). \
            applymap(self.center_vals, subset=['TOTAL_TESTCASES']). \
            applymap(self.center_vals, subset=['TESTCASES_PASSED']). \
            applymap(self.center_vals, subset=['TESTCASES_FAILED']).set_table_styles(
            [{'selector': 'tr.hover td', 'props': [('background-color', 'lightyellow')]},
             {'selector': 'th, td', 'props': [('border', '1px solid black'),
                                              ('padding', '4px'), ('text-align', 'left')]},
             {'selector': 'th', 'props': [('font-family', 'Century Gothic'), ('font-size', '9pt')]},
             {'selector': 'thead', 'props': [('background-color', 'lightblue')]},
             {'selector': '', 'props': [('border-collapse', 'collapse'),
                                        ('border', '1px solid black')]},
             ]).set_properties(**{'font-size': '10pt', 'font-family': 'Century Gothic'}).render()
        html1 = html1.replace('<th class="blank level0 ></th>', '<th class="blank level0">S_NO</th>')

        html_content = html1 + "<b>Automation Execution Detailed Summary:</br></br>" + html
        with open(html_file, 'w') as f:
            f.write(html_content)
        f.close()

    def highlight_vals(self, val, color='lightgreen'):
        if val == 'PASS':
            return 'background-color: %s' % color
        else:
            return 'background-color: red'

    def center_vals(self, val):
        return 'text-align: center'

    def highlight_cols_green(self, val, color='lightgreen'):
        return 'background-color: %s' % color

    def highlight_cols_red(self, val, color='red'):
        return 'background-color: %s' % color

    def left_vals(val):
        return 'text-align: left'



from cqepyutils.PandasUtils import PandasUtils
from cqepyutils.DBUtils import DBUtils
from cqepyutils.RFUtils import RFUtils

# PandasUtils
find_delimiter = PandasUtils().find_delimiter
create_dataframe_from_file = PandasUtils().create_dataframe_from_file
write_df_to_csv = PandasUtils().write_df_to_csv
write_df_to_psv = PandasUtils().write_df_to_psv
df_diff = PandasUtils().df_diff
files_diff = PandasUtils().files_diff
find_common_files = PandasUtils().find_common_files
compare_files = PandasUtils().compare_files
compare_file_contents = PandasUtils().compare_file_contents
create_comparison_df = PandasUtils().create_comparison_df
files_diff_with_diffs = PandasUtils().files_diff_with_diffs
replace_values_in_string = PandasUtils().replace_values_in_string
get_file_format_using_icd = PandasUtils().get_file_format_using_icd
consolidate_execution_summary = PandasUtils().consolidate_execution_summary

# DBUtils
get_query_from_file = DBUtils().get_query_from_file
execute_query = DBUtils().execute_query
visualize_data = DBUtils().visualize_data
execute_sql_query_and_visualize_data = DBUtils().execute_sql_query_and_visualize_data
get_dataframe_columns = DBUtils().get_dataframe_columns



import duckdb
import pandas as pd
import os
import logging
import time


def initialize_logging():
    logging.basicConfig(level=logging.INFO)
    logging.info('Logging initialized')


def connect_to_database(db_path=':memory:'):
    conn = duckdb.connect(database=db_path, read_only=False)
    logging.info('Connected to database')
    return conn


# Main initialization
initialize_logging()
conn = connect_to_database()


def read_fixed_length_file(file_path, structure):
    logging.info(f'Reading fixed-length file: {file_path}')
    start_time = time.time()
    data = []
    with open(file_path, 'r') as file:
        for line in file:
            row = {}
            for start, end, column_name in structure:
                row[column_name] = line[start:end].strip()
            data.append(row)
    logging.info(f'Read fixed-length file completed in {time.time() - start_time} seconds')
    return data


def load_data_into_duckdb(conn, file_path, table_name, mapping_file=None):
    logging.info(f'Loading data into DuckDB from file: {file_path}')
    start_time = time.time()
    extension = os.path.splitext(file_path)[1].lower()
    if extension == '.dat':
        # Read the structure from the Mapping.csv file
        mapping_df = pd.read_csv(mapping_file)
        # Trim spaces from the field_name column
        mapping_df['IGP Base Field'] = mapping_df['IGP Base Field'].str.strip()
        column_names = mapping_df['IGP Base Field'].tolist()
        structure = list(zip(mapping_df["'Start Position'"], mapping_df['End Position'], column_names))
        data = read_fixed_length_file(file_path, structure)
        conn.register(table_name, pd.DataFrame(data, columns=column_names))
    elif extension == '.csv':
        conn.execute(f"CREATE TABLE {table_name} AS SELECT * FROM read_csv_auto('{file_path}')")
    elif extension == '.xml':
        data = pd.read_xml(file_path)
        conn.register(table_name, data)
    elif extension == '.json':
        data = pd.read_json(file_path)
        conn.register(table_name, data)
    else:
        raise ValueError(f"Unsupported file format: {extension}")
    # Return the column names for the loaded table
    logging.info(f'Loading data into DuckDB completed in {time.time() - start_time} seconds')
    column_info = conn.execute(f"PRAGMA table_info({table_name})").fetchall()
    return [col[1] for col in column_info]


def run_data_comparison(conn, src_path, trg_path, primary_keys, table_src='table_src', table_trg='table_trg'):
    logging.info('Running data comparison')
    start_time = time.time()

    # Loading data into DuckDB
    columns_src_names = load_data_into_duckdb(conn, src_path, table_src)
    columns_trg_names = load_data_into_duckdb(conn, trg_path, table_trg)

    # Define keys string for SQL queries
    # keys_str = ' AND '.join([f"a.{key} = b.{key}" for key in PRIMARY_KEYS])
    keys_str = ' AND '.join([f'a."{key}" = b."{key}"' for key in primary_keys])

    # List of columns to compare excluding primary keys
    compare_columns = [col for col in columns_src_names if col not in primary_keys]

    # Data breaks
    start_time = time.time()
    total_data_break_query = f"""
    SELECT a.*, b.*
    FROM {table_src} AS a
    LEFT JOIN {table_trg} AS b ON ({keys_str})
    """
    total_data_break_data = conn.execute(total_data_break_query).fetchall()
    logging.info(f"Total Data Break Query completed in {time.time() - start_time} seconds.")

    # Determine the indices for the source and target columns
    src_indices = [columns_src_names.index(col) for col in compare_columns]
    trg_indices = [columns_trg_names.index(col) + len(columns_src_names) for col in compare_columns]

    # Prepare the data for CSV
    total_data_break = []
    for row in total_data_break_data:
        mismatched_columns = {}
        for key in primary_keys:
            key_index = columns_src_names.index(key)
            mismatched_columns[key] = row[key_index]
        for idx, col in enumerate(compare_columns):
            src_value = row[src_indices[idx]]
            # Check if the target value exists
            if trg_indices[idx] < len(row):
                trg_value = row[trg_indices[idx]]
            else:
                trg_value = None  # Set to None if target value doesn't exist
            # Check for mismatch and capture the column name and differences
            if src_value != trg_value:
                mismatched_columns[f"{col}_Expected"] = src_value
                mismatched_columns[f"{col}_Actual"] = trg_value
        if len(mismatched_columns) > len(primary_keys):
            total_data_break.append(mismatched_columns)

    # Write the data to CSV
    pd.DataFrame(total_data_break).to_csv('total_data_break.csv', index=False)

    # Only in Source
    start_time = time.time()
    only_in_source_query = f"""
    SELECT a.*
    FROM {table_src} AS a
    LEFT JOIN {table_trg} AS b ON ({keys_str})
    WHERE b.{primary_keys[0]} IS NULL
    """
    only_in_source = conn.execute(only_in_source_query).fetchall()
    logging.info(f"Only in Source Query completed in {time.time() - start_time} seconds.")
    pd.DataFrame(only_in_source, columns=columns_src_names).to_csv("only_in_source.csv", index=False)

    # Only in Target
    start_time = time.time()
    only_in_target_query = f"""
    SELECT b.*
    FROM {table_trg} AS b
    LEFT JOIN {table_src} AS a ON ({keys_str})
    WHERE a.{primary_keys[0]} IS NULL
    """
    only_in_target = conn.execute(only_in_target_query).fetchall()
    logging.info(f"Only in Target Query completed in {time.time() - start_time} seconds.")
    pd.DataFrame(only_in_target, columns=columns_trg_names).to_csv("only_in_target.csv", index=False)

    # Compute metrics and write to separate CSVs
    keys_str_no_alias = ', '.join(primary_keys)

    # Source Key Duplicate
    start_time = time.time()
    source_key_duplicate_query = f"""
    SELECT *
    FROM {table_src}
    WHERE ({keys_str_no_alias}) IN
        (SELECT {keys_str_no_alias}
         FROM {table_src}
         GROUP BY ({keys_str_no_alias})
         HAVING COUNT(*) > 1)
    """

    source_key_duplicate = conn.execute(source_key_duplicate_query).fetchall()
    logging.info(f"Source Key Duplicate Query completed in {time.time() - start_time} seconds.")
    pd.DataFrame(source_key_duplicate, columns=columns_src_names).to_csv('source_key_duplicate.csv', index=False)

    # Target Key Duplicate
    start_time = time.time()
    target_key_duplicate_query = f"""
    SELECT *
    FROM {table_trg}
    WHERE ({keys_str_no_alias}) IN
        (SELECT {keys_str_no_alias}
         FROM {table_trg}
         GROUP BY ({keys_str_no_alias})
         HAVING COUNT(*) > 1)
    """
    target_key_duplicate = conn.execute(target_key_duplicate_query).fetchall()
    logging.info(f"Target Key Duplicate Query completed in {time.time() - start_time} seconds.")
    pd.DataFrame(target_key_duplicate, columns=columns_trg_names).to_csv('target_key_duplicate.csv', index=False)

    # Metrics for the summary
    source_total = conn.execute(f"SELECT COUNT(*) FROM {table_src}").fetchone()[0]
    target_total = conn.execute(f"SELECT COUNT(*) FROM {table_trg}").fetchone()[0]
    total_key_matched = \
        conn.execute(f"SELECT COUNT(*) FROM {table_src} AS a JOIN {table_trg} AS b ON ({keys_str})").fetchone()[0]
    only_in_source_count = len(only_in_source)
    only_in_target_count = len(only_in_target)

    # Detect duplicates in source and target based on primary keys
    source_key_duplicate_count = \
        conn.execute(f"SELECT COUNT(DISTINCT {', '.join(primary_keys)}) FROM {table_src}").fetchone()[0]
    target_key_duplicate_count = \
        conn.execute(f"SELECT COUNT(DISTINCT {', '.join(primary_keys)}) FROM {table_trg}").fetchone()[0]

    # Calculate total data breaks
    total_data_break_count = len(total_data_break)

    # Create summary dictionary
    summary_data = {
        'Source Total': source_total,
        'Target Total': target_total,
        'Total Data Breaks': total_data_break_count,
        'Only in Source': only_in_source_count,
        'Only in Target': only_in_target_count,
        'Source Key Duplicate': source_key_duplicate_count,
        'Target Key Duplicate': target_key_duplicate_count
    }

    # Write the summary data to CSV
    pd.DataFrame([summary_data]).to_csv("summary.csv", index=False)

    logging.info(f'Data comparison completed in {time.time() - start_time} seconds')


def main():
    src_path = 'C:/Users/ranji/OneDrive/PycharmProjects/pythontest/src/Org.csv'
    trg_path = 'C:/Users/ranji/OneDrive/PycharmProjects/pythontest/trg/Org.csv'
    primary_keys = ['Organization_Id']
    run_data_comparison(conn, src_path, trg_path, primary_keys)


if __name__ == "__main__":
    main()



import pandas as pd
import plotly.graph_objects as go
import plotly.express as px
from plotly.subplots import make_subplots
from robot.api.deco import keyword
from robot.api import logger
import matplotlib.pyplot as plt
import seaborn as sns


@keyword('Visualize Data')
def visualize_data(*dataframes, x_col: str, y_col: str,
                   plot_type: str = 'bar', titles: list = None, plot_name: str = 'Trace') -> None:
    """
    Visualizes query results as a plot and displays them as a DataFrame.

    Args:
        *dataframes (pd.DataFrame): Variable number of Pandas DataFrames.
        x_col (str): The name of the column to use for the x-axis of the plot.
        y_col (str): The name of the column to use for the y-axis of the plot.
        plot_type (str, optional): The type of plot to use. Default is 'bar', but can be 'line', 'scatter', etc.
        titles (list, optional): A list of titles for each table. Default is None.
        plot_name (str, optional): The name of the trace in the plot. Default is 'Trace'.

    Examples:
        | *Keyword*          | *Arguments*                                                               |
        | `Visualize Data`   | dataframe=x, dataframe2=y, x_col=column_name_1, y_col=column_name_2, plot_type='bar', titles=['Table 1', 'Table 2'], plot_name='Trace' |
        | `Visualize Data`   | dataframe=x, dataframe2=y, x_col=column_name_3, y_col=column_name_4, plot_type='scatter', titles=['Table A', 'Table B'], plot_name='Data' |
        | `Visualize Data`   | dataframe=x, dataframe2=y, x_col=column_name_5, y_col=column_name_6, plot_type='line', titles=['Table X', 'Table Y'], plot_name='Line Data' |
    """

    num_dataframes = len(dataframes)

    if titles is None or len(titles) != num_dataframes:
        titles = ['Table {}'.format(i + 1) for i in range(num_dataframes)]

    logger.info(f'Step - 1: Starting data visualization for {num_dataframes} dataframes.')

    if num_dataframes == 1:
        # If only one dataframe is provided, display it in the center
        fig = go.Figure(data=[go.Table(
            header=dict(values=[col.replace('_', ' ').capitalize() for col in dataframes[0].columns],  # Modify headers
                        fill_color='rgb(30, 53, 76)',
                        font=dict(color='white')),
            cells=dict(values=dataframes[0].T.applymap(lambda x: f'<b>{x}</b>').values.tolist(),  # Center-align values
                       fill_color=[[['#f2f2e8', '#ffffff'][i % 2] for i in range(len(dataframes[0]))]],
                       font=dict(color='black'))
        )])
        fig.update_layout(title=titles[0])
    else:
        # If multiple dataframes are provided, create subplots in a grid layout
        num_rows = (num_dataframes + 1) // 2
        fig = make_subplots(rows=num_rows, cols=2, specs=[[{'type': 'domain'}, {'type': 'domain'}]] * num_rows)

        for i, (dataframe, title) in enumerate(zip(dataframes, titles)):
            row = (i // 2) + 1
            col = (i % 2) + 1

            table = go.Table(
                header=dict(values=[col.replace('_', ' ').capitalize() for col in dataframe.columns],  # Modify headers
                            fill_color='rgb(30, 53, 76)',
                            font=dict(color='white')),
                cells=dict(values=dataframe.T.applymap(lambda x: f'<b>{x}</b>').values.tolist(),  # Center-align values
                           fill_color=[[['#f2f2e8', '#ffffff'][i % 2] for i in range(len(dataframe))]],
                           font=dict(color='black'))
            )

            fig.add_trace(table, row=row, col=col)
            fig.update_xaxes(title_text=title, row=row, col=col)  # Update x-axis title
            fig.update_yaxes(title_text='', row=row, col=col)  # Remove y-axis title

    # Update layout
    fig.update_layout(
        height=500 * ((num_dataframes + 1) // 2) if num_dataframes > 1 else 500,
        margin=dict(t=50, l=50, r=50),
    )
    fig.show()
    logger.info(f'Step - 2: Data visualization complete for {num_dataframes} dataframes.')


@keyword('Visualize Combined Data')
def visualize_data_combined(dataframe1: pd.DataFrame, dataframe2: pd.DataFrame, x_col1: str, y_col1: str, x_col2: str, y_col2: str) -> None:
    """
    Visualizes query results as a plot and displays them as a DataFrame.

    Args:
        dataframe1 (pd.DataFrame): First Pandas DataFrame.
        dataframe2 (pd.DataFrame): Second Pandas DataFrame.
        x_col1 (str): The name of the column to use for the x-axis of the first Seaborn plot.
        y_col1 (str): The name of the column to use for the y-axis of the first Seaborn plot.
        x_col2 (str): The name of the column to use for the x-axis of the second Seaborn plot.
        y_col2 (str): The name of the column to use for the y-axis of the second Seaborn plot.

    Examples:
        | *Keyword*                | *Arguments*                                                                   |
        | `Visualize Combined Data` | dataframe1=x, dataframe2=y, x_col1=column_name_1, y_col1=column_name_2, x_col2=column_name_3, y_col2=column_name_4 |
    """

    fig = make_subplots(
        rows=2, cols=2,
        specs=[[{'type': 'xy'}, {'type': 'xy'}],
               [{'type': 'table', 'colspan': 2}, None]],
        subplot_titles=('Seaborn Chart 1', 'Seaborn Chart 2', 'Dataframe 1', 'Dataframe 2')
    )

    # Generate Seaborn plots and add them to the first row
    plt.figure(figsize=(8, 6))
    sns.barplot(x=x_col1, y=y_col1, data=dataframe1)
    ax1 = plt.gca()
    ax1.get_xaxis().set_visible(False)
    ax1.get_yaxis().set_visible(False)
    fig.add_trace(go.Scatter(x=[None], y=[None], mode='markers'), row=1, col=1)  # Placeholder for Seaborn chart
    plt.close()

    plt.figure(figsize=(8, 6))
    sns.barplot(x=x_col2, y=y_col2, data=dataframe2)
    ax2 = plt.gca()
    ax2.get_xaxis().set_visible(False)
    ax2.get_yaxis().set_visible(False)
    fig.add_trace(go.Scatter(x=[None], y=[None], mode='markers'), row=1, col=2)  # Placeholder for Seaborn chart
    plt.close()

    # Convert data frames to Plotly tables and add them to the second row
    table_trace1 = go.Table(
        header=dict(values=list(dataframe1.columns), fill_color='rgb(30, 53, 76)', font=dict(color='white')),
        cells=dict(values=dataframe1.T.applymap(lambda x: f'<b>{x}</b>').values.tolist(),
                   fill_color=[[['#f2f2e8', '#ffffff'][i % 2] for i in range(len(dataframe1))]],
                   font=dict(color='black'))
    )

    table_trace2 = go.Table(
        header=dict(values=list(dataframe2.columns), fill_color='rgb(30, 53, 76)', font=dict(color='white')),
        cells=dict(values=dataframe2.T.applymap(lambda x: f'<b>{x}</b>').values.tolist(),
                   fill_color=[[['#f2f2e8', '#ffffff'][i % 2] for i in range(len(dataframe2))]],
                   font=dict(color='black'))
    )

    fig.add_trace(table_trace1, row=2, col=1)
    fig.add_trace(table_trace2, row=2, col=2)

    # Update layout
    fig.update_layout(
        height=800,
        title_text="Combined Data Visualization",
        showlegend=False
    )

    # Show the figure
    fig.show()



import os
import filecmp
import difflib
import csv
from typing import List, Tuple
from typing import Union
import pandas as pd
import numpy as np
from robot.api import logger
from robot.api.deco import keyword
from cqepyutils.DBUtils import DBUtils


class PandasUtils:

    def __init__(self):
        pass

    @keyword('Find Delimiter In File')
    def find_delimiter(self, filename: str) -> Union[str, None]:
        """
        Find the delimiter used in a CSV file.

        `Args:`
        | *`Name`* | *`Type`* | *`Description`* |
        | `filename`|  The name of the CSV file.

        `Returns:`
        | *`Type`* | *`Description`* |
        | `str` | `The delimiter used in the CSV file as a string.` |

        Examples:
        | ${delimiter} | Find Delimiter | example.csv |
        """
        _, ext = os.path.splitext(filename)

        if ext in ['.xls', '.xlsx']:
            logger.info(f'{filename} is an Excel file. Cannot find delimiter.')
            return None

        try:
            with open(filename, 'rb') as f:
                # Read a sample of the data to analyze the dialect
                sample = f.read(1024).decode()

                # List of delimiters to test
                delimiters = [',', '\t', ';', '|', ':', '~', '\x01']
                delimiter_str = "".join(delimiters)

                # Use csv.Sniffer to guess the dialect
                dialect = csv.Sniffer().sniff(sample, delimiters=delimiter_str)

            delimiter = chr(getattr(dialect, 'delimiter'))

            # Log the delimiter
            logger.info(f'The delimiter in {filename} is "{delimiter}"')

            return delimiter
        except UnicodeDecodeError:
            logger.info(f'{filename} is not a delimiter file')
            return None

    @keyword(name="Create Dataframe From File")
    def create_dataframe_from_file(self, file_path: str, delimiter: str = None, has_header: bool = True,
                                   width: list = None, colspecs: list = None,
                                   column_names: list = None, dtypes: dict = None,
                                   encoding: str = 'ISO-8859-1',
                                   on_bad_lines: str = 'warn', skiprows: int = 0, skipfooter: int = 0):
        """
        Creates a Pandas DataFrame from a CSV, PSV, or fixed-width file.

        `Args:`
        | *`Name`* | *`Type`* | *`Description`* |
        | `file_path`    | `(str)` | `The path to the input file.` |
        | `delimiter`    | `(str)` | The delimiter character used in the input file. Default is ','.` |
        | `has_header`   | `(bool)` | `Whether the input file has headers. Default is True.` |
        | `width`        | `(list or tuple)` | `A list or tuple of integers specifying the width of each fixed-width field. Required when reading a fixed-width file. Default is None.` |
        | `encoding`     | `(str)` | `The encoding of the input file. Default is 'ISO-8859-1'.` |
        | `on_bad_lines` | `(str)` | `What to do with bad lines encountered in the input file. Valid values are 'raise', 'warn', and 'skip'. Default is 'warn'.` |
        | `skiprows`     | `(int or list-like)` | `Line numbers to skip (0-indexed) or number of lines to skip (int) at the start of the file. Default is 0.` |
        | `skipfooter`   | `(int)` | `Number of lines to skip at the end of the file. Default is 0.` |

        `Returns:`
        | *`Type`* | *`Description`* |
        | `pandas.DataFrame` | `The DataFrame created from the input file.` |

        Examples:
        | Create Dataframe From File | /path/to/file.csv |
        | Create Dataframe From File | /path/to/file.psv | delimiter='|', has_header=False |
        | Create Dataframe From File | /path/to/file.xlsx | has_header=False |
        | Create Dataframe From File | /path/to/file.fwf | width=[10, 20, 30] |
        | Create Dataframe From File | /path/to/file.csv | encoding='utf-8', on_bad_lines='raise' |
        """

        # Set default dtypes as 'str' if not provided
        if dtypes is None:
            dtypes = 'str'

        # Determine if the file has a delimiter based on the provided delimiter
        has_delimiter = delimiter is not None and delimiter != ''

        # Determine the file type based on the file extension
        file_ext = file_path.split('.')[-1].lower()
        if has_delimiter:
            # Log info message
            logger.info(
                f"        Reading delimited file '{file_path}' with delimiter '{delimiter}' and encoding '{encoding}'")
            # Read delimited file into a DataFrame
            df = pd.read_csv(file_path, delimiter=delimiter, header=None if not has_header else 'infer',
                             names=column_names, encoding=encoding, on_bad_lines=on_bad_lines,
                             skiprows=skiprows, skipfooter=skipfooter, dtype=dtypes, index_col=False, engine='python')
        elif file_ext == 'xlsx':
            # Log info message
            logger.info(f"        Reading Excel file '{file_path}'")
            # Read Excel file into a DataFrame
            df = pd.read_excel(file_path, skiprows=skiprows, skipfooter=skipfooter, dtype=dtypes)
        elif file_ext == 'dat':
            if width is not None:
                # Log info message
                logger.info(f"        Reading fixed-width file '{file_path}' with specified width")
                # Read fixed-width file into a DataFrame using width parameter
                df = pd.read_fwf(file_path, widths=width, header=None if not has_header else 'infer',
                                 names=column_names,
                                 encoding=encoding, on_bad_lines=on_bad_lines,
                                 skiprows=skiprows, skipfooter=skipfooter, dtype=dtypes, index_col=False)
            elif colspecs is not None:
                # Log info message
                logger.info(f"        Reading fixed-width file '{file_path}' with specified colspecs")
                # Read fixed-width file into a DataFrame using colspecs parameter
                df = pd.read_fwf(file_path, colspecs=colspecs, header=None if not has_header else 'infer',
                                 names=column_names, encoding=encoding, on_bad_lines=on_bad_lines,
                                 skiprows=skiprows, skipfooter=skipfooter, dtype=dtypes, index_col=False)
            else:
                # Log error message and raise exception
                logger.error("Error: For '.dat' files, you must provide either 'width' or 'colspecs'.")
                raise Exception("Invalid '.dat' file configuration")
        else:
            # Log error message and raise exception
            logger.error(f"Error: Unsupported file type '{file_ext}'. "
                         f"Supported file types are 'csv', 'psv', 'xlsx', and 'dat'.")
            raise Exception(f"Unsupported file type '{file_ext}'")

        # Log info message
        logger.info(f"        DataFrame created with {df.shape[0]} rows and {df.shape[1]} columns")
        # Return the DataFrame
        return df

    @keyword("Write DataFrame To CSV File")
    def write_df_to_csv(self, df_to_write: pd.DataFrame, file_path: str, file_name: str, index: bool = False):
        """
        Write the DataFrame to a CSV file.

        `Args:`
        | *`Name`* | *`Type`* | *`Description`* |
        | `df_to_write`    | `pandas.DataFrame` | `The DataFrame to write to the CSV file.` |
        | `file_path`    | `str` | `The path where the CSV file needs to be created.` |
        | `file_name`    | `str` | `The name of the CSV file to be created.` |
        | `index`    | `bool` | `Whether to include the index in the CSV file. Default is False.` |

        `Returns:`
        | *`Type`* | *`Description`* |
        | `None`    | `This method does not return anything.` |

        Examples:
        | Write DataFrame To CSV File | ${df} | /path/to/file/ | example.csv |
        | Write DataFrame To CSV File | ${df} | /path/to/file/ | example.csv | True |
        """
        logger.info('Step 1: Writing DataFrame to CSV file...')
        try:
            df_to_write.to_csv(path_or_buf=file_path + '/' + file_name, mode='w', index=index)
            logger.info('Step 2: Writing DataFrame to CSV file completed successfully.')
        except Exception as e:
            logger.error(f'Step 2: Writing DataFrame to CSV file failed with error: {e}')

    @keyword("Write DataFrame To PSV File")
    def write_df_to_psv(self, df_to_write: pd.DataFrame, file_path: str, file_name: str):
        """
        Write the DataFrame to a PSV (pipe-separated values) file.

        `Args:`
        | *`Name`* | *`Type`* | *`Description`* |
        | `df_to_write` | `pandas.DataFrame` | `The DataFrame to write to the PSV file.` |
        | `file_path` | `str` | `The path where the PSV file needs to be created.` |
        | `file_name` | `str` | `The name of the PSV file to be created.` |

        `Returns:`
        | *`Type`* | *`Description`* |
        | `None` | `This method does not return anything.` |

        Examples:
        | Write DataFrame To PSV File | ${df} | /path/to/file/ | example.psv |
        """

        logger.info('Step 1: Writing DataFrame to PSV file...')
        try:
            df_to_write.to_csv(path_or_buf=file_path + '/' + file_name, mode='w', sep='|', index=False)
            logger.info('Step 2: Writing DataFrame to PSV file completed successfully.')
        except Exception as e:
            logger.error(f'Step 2: Writing DataFrame to PSV file failed with error: {e}')

    @keyword("Compare All File Contents In Directories")
    def df_diff(self, source_file_path_name: str, target_file_path_name: str, delimiter: str = None,
                has_header: bool = True,
                width: list = None, colspecs: list = None, column_names: list = None, dtypes: dict = None,
                encoding: str = 'ISO-8859-1', on_bad_lines: str = 'warn', skiprows: int = 0, skipfooter: int = 0,
                key_columns: list = None, ignore_columns: list = None, report_type: str = 'all'):
        """
        Compare the difference between two files' contents and generate a detailed mismatch report.

        `Args:`
        | *`Name`* | *`Type`* | *`Description`* |
        | `source_file_path_name` | `(str)` | `The path to the first file containing the source DataFrame to compare.` |
        | `target_file_path_name` | `(str)`  | `The path to the second file containing the target DataFrame to compare.` |
        | `delimiter` | `(str)` | `The delimiter character used in the input files. Default is ','.` |
        | `has_header` | `(bool)` | `Whether the input files have headers. Default is True.` |
        | `width` | `(list or tuple)` | `A list or tuple of integers specifying the width of each fixed-width field. Required when reading a fixed-width file. Default is None.` |
        | `colspecs` | `(list or tuple)` | `A list or tuple of pairs (int, int) specifying the column ranges for fixed-width fields. Required when reading a fixed-width file. Default is None.` |
        | `column_names` | `(list)` | `A list of column names to use if the input files do not have headers.` |
        | `dtypes` | `(dict)` | `A dictionary specifying the data types of columns in the DataFrame.` |
        | `encoding` | `(str)` | `The encoding of the input files. Default is 'ISO-8859-1'.` |
        | `on_bad_lines` | `(str)` | `What to do with bad lines encountered in the input files. Valid values are 'raise', 'warn', and 'skip'. Default is 'warn'.` |
        | `skiprows` | `(int or list-like)` | `Line numbers to skip (0-indexed) or number of lines to skip (int) at the start of the files. Default is 0.` |
        | `skipfooter` | `(int)` | `Number of lines to skip at the end of the files. Default is 0.` |
        | `key_columns` | `(list)` | `A list of column names to use as the primary keys for the comparison.` |
        | `ignore_columns` | `(list)` | `A list of column names to ignore during the comparison.` |
        | `report_type` | `(str)` | `The type of report to generate. Valid values are 'all', 'duplicate', or 'key_mismatch'. Default is 'all'.` |

        `Returns:`
        | *`Type`* | *`Description`* |
        | `pandas.DataFrame` | `A DataFrame containing the rows that are different between the two input DataFrames.` |

        Examples:
        | ${source_df} | Read CSV | path/to/source.csv |
        | ${target_df} | Read CSV | path/to/target.csv |
        | ${diff} | Compare All File Contents In Directories | ${source_df} | ${target_df} | key_columns=['col1', 'col2'] | ignore_columns=['col3'] |
        """

        logger.info(
            '****************************************************************************************************')
        logger.info('PandasUtil Data Frame Comparison - Cell by Cell comparison with detailed mismatch report')
        logger.info(
            '****************************************************************************************************')
        logger.info('Step-01 : Based on file format create the data frames with delimiter(sep)')

        source_df = self.create_dataframe_from_file(source_file_path_name, delimiter, has_header, width, colspecs,
                                                    column_names,
                                                    dtypes, encoding, on_bad_lines, skiprows, skipfooter)
        target_df = self.create_dataframe_from_file(target_file_path_name, delimiter, has_header, width, colspecs,
                                                    column_names,
                                                    dtypes, encoding, on_bad_lines, skiprows, skipfooter)

        # Store total records in source and target df
        total_target = round(len(source_df))
        total_source = round(len(target_df))
        total_mismatch = total_target - total_source

        logger.info('Step-02 : Remove the columns based on ignore columns list')

        # If ignore columns are specified, remove those columns from comparison
        if ignore_columns:
            for col in ignore_columns:
                # Check if the column exists in source_df and target_df and remove it
                if col in source_df.columns:
                    source_df.drop(columns=col, inplace=True)
                if col in target_df.columns:
                    target_df.drop(columns=col, inplace=True)

                # If col is an integer within valid range, remove by index
                elif isinstance(col, int) and 0 <= col < len(source_df.columns):
                    source_df.drop(columns=source_df.columns[col], inplace=True)

                elif isinstance(col, int) and 0 <= col < len(target_df.columns):
                    target_df.drop(columns=target_df.columns[col], inplace=True)

        logger.info('Step-03 : Check for duplicate rows in both source and target')

        # Check for column differences in source_df and target_df
        source_df_col_diff = set(source_df.columns) - set(target_df.columns)
        target_df_col_diff = set(target_df.columns) - set(source_df.columns)

        logger.debug(source_df_col_diff)
        source_df_col_diff = set(source_df.columns) - set(target_df.columns)

        # If key column is not specified then consider all columns except last column
        if len(key_columns) == 0:
            key_columns = source_df.columns.tolist()
            key_columns.pop()

        # Sort both target and source data frame
        source_df.sort_values(by=key_columns, ascending=True, inplace=True)
        target_df.sort_values(by=key_columns, ascending=True, inplace=True)

        # Check for duplicate key columns in target and source data frame
        source_df_dup_df = source_df[source_df[key_columns].duplicated()]
        target_df_dup_df = target_df[target_df[key_columns].duplicated()]

        logger.debug(source_df_dup_df)
        logger.debug(target_df_dup_df)
        logger.debug(len(source_df_dup_df))
        logger.debug(len(target_df_dup_df))

        total_target_dup = round(len(source_df_dup_df))
        total_source_dup = round(len(target_df_dup_df))

        logger.info('Step-04 : Remove duplicate records from source and target')

        # Get the duplicate key columns
        dup_target_df = source_df_dup_df.copy()
        dup_source_df = target_df_dup_df.copy()
        dup_target_df['file'] = 'target'
        dup_source_df['file'] = 'source'

        # Combine the duplicate keys from target and source data frame
        dup_cons_df = pd.concat([dup_target_df, dup_source_df], axis=0)
        dup_cons_df.reset_index(inplace=True)
        dup_cons_df.drop('index', axis=1, inplace=True)

        # Drop the duplicate columns before detailed comparison
        source_df.drop_duplicates(key_columns, inplace=True)
        target_df.drop_duplicates(key_columns, inplace=True)

        logger.debug(dup_target_df)
        logger.debug(dup_source_df)
        logger.debug(dup_cons_df)

        logger.info('Step-05 : Sort the source and target based on key columns and reset the index')

        # Sort source_df and target_df based on key columns and reset the index
        source_df.sort_values(by=key_columns, ascending=True, inplace=True)
        target_df.sort_values(by=key_columns, ascending=True, inplace=True)
        source_df.reset_index(inplace=True)
        target_df.reset_index(inplace=True)

        # Set the index based on key columns in source_df and target_df. Remove the default index column
        source_df = source_df.set_index(key_columns, drop=True, append=False, inplace=False, verify_integrity=True)
        target_df = target_df.set_index(key_columns, drop=True, append=False, inplace=False, verify_integrity=True)
        source_df = source_df.drop('index', axis=1)
        target_df = target_df.drop('index', axis=1)

        logger.info('Step-06 : Identify the rows matching based on key in both source and target')

        # Identify the rows matching based on key in both source_df and target_df
        merge_outer_df = pd.merge(source_df, target_df, how='outer', left_index=True, right_index=True,
                                  indicator='source')
        # merge_outer_df = pd.merge(source_df_key_columns, target_df_key_columns, how='outer', on=key_columns,
        # indicator='source')

        # Based on the key columns create key matched and mismatched details
        key_matched_df = merge_outer_df.loc[merge_outer_df['source'] == 'both'].copy()
        logger.debug(len(key_matched_df))
        key_mismatched_df = merge_outer_df.loc[merge_outer_df['source'] != 'both'].copy()
        key_mismatched_df = key_mismatched_df[['source']]

        # key_matched_df['source'] = 'Matched'
        # key_mismatched_df['source'] = 'MisMatched'
        logger.debug(key_mismatched_df)

        # Update the source column left_only to source and right_only to target
        # key_mismatched_df.loc[key_mismatched_df['source'] == 'left_only', 'source'] = 'source'

        target_key_mismatch = len(key_mismatched_df[key_mismatched_df.source == 'left_only'])
        source_key_mismatch = len(key_mismatched_df[key_mismatched_df.source == 'right_only'])

        # Create DataFrames for Data Only In Source and Data Only In Target
        data_only_in_source_df = source_df[~source_df.index.isin(target_df.index)].copy()
        data_only_in_target_df = target_df[~target_df.index.isin(source_df.index)].copy()

        logger.info('Step-07 : Create the summary report based on count diff, duplicate rows and key mismatches')

        only_in_source_count = len(data_only_in_source_df)
        only_in_target_count = len(data_only_in_target_df)

        # Extract source and target file names from file paths
        source_file_name = os.path.basename(source_file_path_name)
        target_file_name = os.path.basename(target_file_path_name)

        # Calculate count details for each column
        count_details = {
            'Source File Name': [source_file_name],  # Add source file name
            'Target File Name': [target_file_name],  # Add target file name
            'Source Total': [len(source_df)],
            'Target Total': [len(target_df)],
            'Total Key Matched': [len(key_matched_df)],
            'Total Keys & Data Matched': [0],
            'Total Keys & Data MisMatched': [0],
            'Total Data Breaks': [0],
            'Only in Source': [only_in_source_count],
            'Only in Target': [only_in_target_count],
            'Source Key Duplicates': [total_source_dup],
            'Target Key Duplicates': [total_target_dup],
            'Key Columns': [', '.join(key_columns)]  # Convert key_columns list to a comma-separated string
        }

        # Create the executive summary df
        exec_summary_df = pd.DataFrame.from_dict(count_details)
        logger.debug(exec_summary_df)

        logger.info('Step-08 : Remove the mismatched key values and proceed further in validation')

        source_df.drop(key_mismatched_df.loc[key_mismatched_df['source'] == 'left_only'].index, inplace=True)
        target_df.drop(key_mismatched_df.loc[key_mismatched_df['source'] == 'right_only'].index, inplace=True)

        # Step-08 to Step-12: Handle different report types
        if report_type == 'all':
            # Continue with all the comparison steps and generate all reports
            pass
        elif report_type == 'duplicate':
            # Return only the duplicate records report and exit the function
            logger.info('Step-09 : Comparison completed and generated info for reports(summary, duplicate')
            logger.info(
                '************************************************************************************************')
            return exec_summary_df, dup_cons_df
        elif report_type == 'key_mismatch':
            # Return only the key mismatch report and exit the function
            logger.info('Step-09 : Comparison completed and generated info for reports(summary, key_mismatch')
            logger.info(
                '************************************************************************************************')
            return exec_summary_df, key_mismatched_df, data_only_in_source_df, data_only_in_target_df
        else:
            # Invalid report type specified, raise an exception
            raise ValueError("Invalid 'report_type' specified. Valid values are 'all', 'duplicate', or 'key_mismatch'.")

        # Check if source_df and target_df are equal
        if source_df.equals(target_df) and only_in_source_count == 0 and only_in_target_count == 0:
            logger.info('No differences found between the DataFrames.')
            logger.info(
                '************************************************************************************************')
            data_only_in_source_df = pd.DataFrame([])
            data_only_in_target_df = pd.DataFrame([])
            cell_comp_df = pd.DataFrame([])
            exec_summary_df.loc[0, 'Total Keys & Data Matched'] = len(source_df)

            return exec_summary_df, dup_cons_df, key_matched_df, key_mismatched_df, data_only_in_source_df, \
                data_only_in_target_df, cell_comp_df

        logger.info('Step-09 : Started cell by cell comparison for key values that exist in both source and target')

        # Verify if columns in both source_df and target_df are same
        if not source_df.columns.equals(target_df.columns):
            logger.debug('Failed - Column mismatch determined')
            # Handle the column mismatch case here if needed
            # ...

        logger.info('Step-10 : Verify column data types in both the files, if not convert based on source')
        if not source_df.dtypes.equals(target_df.dtypes):
            logger.debug('Data Types are different, trying to convert')
            target_df = target_df.astype(source_df.dtypes)

        logger.info('Step-11 : Verify cell by cell data in both the data frame and generate mismatch report')

        # Identify where cells are different and generate a boolean mask
        source_df.fillna('Empty Cell', inplace=True)
        target_df.fillna('Empty Cell', inplace=True)
        diff_mask = (source_df != target_df)  # & ~(source_df.isnull() & target_df.isnull())

        # Special case: Treat 0 values in both DataFrames as matches
        zero_mask = (source_df == 0) & (target_df == 0)
        diff_mask = diff_mask | zero_mask

        # Check if diff_mask is empty (no differences found)
        if len(diff_mask) == 0:
            logger.info('Step-12 : No differences found between the DataFrames.')
            cell_comp_df = pd.DataFrame(
                columns=key_columns + ['Column', 'Target_Data', 'Source_Data', 'Compare_Result'])
        else:
            # Create a DataFrame with mismatched cells
            cell_comp_df = source_df.where(diff_mask).stack().reset_index()
            key_columns_names = key_columns if isinstance(key_columns, list) else [key_columns]
            cell_comp_df.columns = key_columns_names + ['Column', 'Source_Data']

            def get_index_value(row):
                if len(key_columns_names) > 1:
                    return tuple(row[key] for key in key_columns_names)
                else:
                    return row[key_columns_names[0]]

            # Apply the helper function to get index values
            cell_comp_df['Index_Value'] = cell_comp_df.apply(get_index_value, axis=1)

            # Use Index_Value to fetch Source_Data
            cell_comp_df['Target_Data'] = cell_comp_df.apply(
                lambda row: target_df.at[row['Index_Value'], row['Column']],
                axis=1
            )

            # Add a 'Compare_Result' column
            cell_comp_df['Compare_Result'] = np.where(cell_comp_df['Target_Data'] != cell_comp_df['Source_Data'],
                                                      'Mismatch', 'Match')
            # Assuming key_matched_df and cell_comp_df have the same index, you can get the unique index values
            unique_indices_in_key_matched = set(key_matched_df.index)
            unique_indices_in_cell_comp = set(cell_comp_df['Index_Value'])

            # Find the unique indices in key_matched_df that are not present in cell_comp_df
            matched_indices = unique_indices_in_key_matched.difference(unique_indices_in_cell_comp)

            # Drop the Index_Value column before returning
            cell_comp_df = cell_comp_df.drop(columns=['Index_Value'])

            # Calculate "Total Keys & Data MisMatched"
            total_keys_data_mismatch = int(len(key_matched_df)) - int(len(matched_indices))

            # Calculate "Total Data Breaks"
            total_data_breaks = int(len(cell_comp_df))

            # Calculate the count of unique keys that are matched in both key_matched_df and cell_comp_df
            total_key_data_matched = len(matched_indices)

            # Update the count details in the executive summary DataFrame
            exec_summary_df.loc[0, 'Total Keys & Data Matched'] = total_key_data_matched
            exec_summary_df.loc[0, 'Total Keys & Data MisMatched'] = total_keys_data_mismatch
            exec_summary_df.loc[0, 'Total Data Breaks'] = total_data_breaks

        logger.info(
            'Step-12 : Comparison completed and generated info for reports(summary, keys mismatch, cell by cell)')
        logger.info(
            '****************************************************************************************************')

        return exec_summary_df, dup_cons_df, key_matched_df, key_mismatched_df, data_only_in_source_df, \
            data_only_in_target_df, cell_comp_df

    @keyword('Compare All Files In Directories')
    def files_diff(self, src_dir: str, tgt_dir: str) -> pd.DataFrame:
        """
        Compare all files in source and target directories and return a DataFrame with the comparison results.

        `Args:`
        | *`Name`* | *`Type`* | *`Description`* |
        | `src_dir` | `str` | `The source directory to compare.` |
        | `tgt_dir` | `str` | `TThe target directory to compare.` |

        `Returns:`
        | *`Type`* | *`Description`* |
        | `pd.DataFrame` | `A DataFrame containing the comparison results, including file name, status, comments, and size in both directories.` |

        Examples:
        | ${result}= | Files Diff | ${source_dir} | ${target_dir} |
        | ${result}= | Files Diff | /path/to/source/dir | /path/to/target/dir |
        """

        logger.info(f"Step 1: Log the comparison directories")
        logger.info(f"Comparing files in {src_dir} and {tgt_dir}")

        common_files = self.find_common_files(src_dir, tgt_dir)

        match, mismatch, errors = self.compare_files(src_dir, tgt_dir, common_files)

        only_in_src = self.find_files_only_in_dir(src_dir, common_files)

        only_in_tgt = self.find_files_only_in_dir(tgt_dir, common_files)

        df = self.create_comparison_df(match, mismatch, errors, only_in_src, only_in_tgt, src_dir, tgt_dir)

        logger.info(f"Step 10: Created File Comparison Report - {len(df)} Files")
        return df

    @keyword('Find Common Files In Directories')
    def find_common_files(self, src_dir: str, tgt_dir: str) -> List[str]:
        """
        Find files that exist in both directories.

        `Args:`
        | *`Name`* | *`Type`* | *`Description`* |
        | `src_dir` | `str` | `The source directory.` |
        | `tgt_dir` | `str` | `The target directory.` |

        `Returns:`
        | *`Type`* | *`Description`* |
        | `List[Tuple[str, str]]` | `A list of tuples containing file paths, one for each file that exists in both directories.` |

        Examples:
        | ${result}= | Find Common Files | ${source_dir} | ${target_dir} |
        | ${result}= | Find Common Files | /path/to/source/dir | /path/to/target/dir |
        """
        logger.info(f"Step 2: Find common files in source and target directories")
        common_files = []
        for file in os.listdir(src_dir):
            if file in os.listdir(tgt_dir):
                common_files.append(file)
        return common_files

    def compare_files(self, src_dir: str, tgt_dir: str, common_files: List[str]) -> Tuple[List[Tuple[str, int, int]],
    List[Tuple[str, int, int, float]],
    List[Tuple[str, str]]]:
        """
        Compare the files in the source and target directories.

        `Args:`
        | *`Name`* | *`Type`* | *`Description`* |
        | `src_dir` | `str` | `The source directory.` |
        | `tgt_dir` | `str` | `The target directory.` |
        | `files` | `list` | `A list of files to compare.` |

        `Returns:`
        | *`Type`* | *`Description`* |
        | `tuple` | `A tuple containing lists of matching files, mismatching files with percentage difference, and files with errors.` |

        Examples:
        | ${src_dir} | Set Variable | /path/to/source |
        | ${tgt_dir} | Set Variable | /path/to/target |
        | @{common_files} | Create List | file1.txt | file2.txt |
        | ${match_files} | ${mismatch_files} | ${error_files} | Compare Files | ${src_dir} | ${tgt_dir} | @{common_files} |
        | Log List | ${match_files} |
        | Log List | ${mismatch_files} |
        | Log List | ${error_files} |

        """
        match = []
        mismatch = []
        errors = []

        logger.info(f"Step 3: Compare each common file")
        for file in common_files:
            try:
                src_path = os.path.join(src_dir, file)
                tgt_path = os.path.join(tgt_dir, file)

                src_size = os.path.getsize(src_path)
                tgt_size = os.path.getsize(tgt_path)

                result = filecmp.cmpfiles(src_dir, tgt_dir, [file], shallow=False)
                if result[0]:
                    match.append((file, src_size, tgt_size))
                else:
                    diff_percentage = self.compare_file_contents(src_path, tgt_path)
                    mismatch.append((file, src_size, tgt_size, diff_percentage))
            except Exception as e:
                errors.append((file, str(e)))

        return match, mismatch, errors

    @keyword('Compare Files Using Diff')
    def compare_file_contents(self, src_path: str, tgt_path: str) -> float:
        """
        Compares file contents using difflib and returns the percentage difference.

        Args:
            src_path (str): The source file path to compare.
            tgt_path (str): The target file path to compare.

        Returns:
            float: The percentage difference between the two files.

        Robot Framework Usage:
        | ${diff_percentage} | Compare File Contents | ${src_path} | ${tgt_path} |

        Examples:
        | ${diff_percentage} | Compare File Contents | /path/to/source/file.txt | /path/to/target/file.txt |
        """
        logger.info(f"Step 4: Compare file contents for {src_path} and {tgt_path}")
        with open(src_path, 'r') as f1, open(tgt_path, 'r') as f2:
            src_lines = f1.readlines()
            tgt_lines = f2.readlines()

        matcher = difflib.SequenceMatcher(None, src_lines, tgt_lines)
        diff_percentage = 100.0 - matcher.ratio() * 100.0

        return round(diff_percentage, 2)

    def find_files_only_in_dir(self, directory: str, common_files: List[str]) -> List[str]:
        """
        Finds files that exist only in a directory and not in the common files list.

        Args:
            directory (str): The directory to search.
            common_files (List[str]): A list of common files.

        Returns:
            List[str]: A list of files that exist only in the directory.

        Robot Framework Usage:
            | *Arguments* |              |
            | directory   | ${dir_path}  |
            | common_files| @{common_files}|
            |             |              |
            | *Returns*   |              |
            |             | @{only_in_dir}|

        Examples:
            | ${common_files}= |  find common files | /path/to/source | /path/to/target |
            | ${only_in_source}= | find files only in dir | /path/to/source | ${common_files} |
            | ${only_in_target}= | find files only in dir | /path/to/target | ${common_files} |
        """
        logger.info(f"Step 4: Find files that exist only in {directory}")
        only_in_dir = []
        for file in os.listdir(directory):
            if file not in common_files:
                only_in_dir.append(file)
        return only_in_dir

    @keyword("Create Comparison Dataframe")
    # @keyword("Create Comparison Dataframe")
    def create_comparison_df(self, match: List[Tuple[str, int, int]],
                             mismatch: List[Tuple[str, int, int, float]],
                             errors: List[Tuple[str, str]],
                             only_in_src: List[str],
                             only_in_tgt: List[str],
                             src_dir: str,
                             tgt_dir: str) -> pd.DataFrame:
        """
        Creates a DataFrame containing the comparison results.

        Args:
            match (List[Tuple[str, int, int]]): A list of matching files and their sizes.
            mismatch (List[Tuple[str, int, int, float]]): A list of mismatching files, their sizes, and the percentage difference.
            errors (List[Tuple[str, str]]): A list of files with errors and the error message.
            only_in_src (List[str]): A list of files only in the source directory.
            only_in_tgt (List[str]): A list of files only in the target directory.
            src_dir (str): Path to the source directory.
            tgt_dir (str): Path to the target directory.

        Returns:
            pd.DataFrame: A DataFrame containing the comparison results, including file name, status, comments, and size in both directories.
        """
        logger.info(f"Step 4: Create comparison DataFrame")
        data = []

        # Add matching files
        for file, src_size, tgt_size in match:
            data.append({"FileName": file,
                         "Status": "MATCH",
                         "Comments": "",
                         "Common_File": "Yes",
                         "Src_Size": src_size,
                         "Tgt_Size": tgt_size,
                         "Diff_Percentage": ""})

        # Add mismatching files
        for file, src_size, tgt_size, diff_percentage in mismatch:
            data.append({"FileName": file,
                         "Status": "MISMATCH",
                         "Comments": f"Percentage Difference: {diff_percentage:.2f}%",
                         "Common_File": "Yes",
                         "Src_Size": src_size,
                         "Tgt_Size": tgt_size,
                         "Diff_Percentage": diff_percentage})

        # Add files with errors
        for file, error in errors:
            data.append({"FileName": file,
                         "Status": "ERROR",
                         "Comments": error,
                         "Common_File": "Yes",
                         "Src_Size": "",
                         "Tgt_Size": "",
                         "Diff_Percentage": ""})

        # Add files only in source directory
        for file in only_in_src:
            data.append({"FileName": file,
                         "Status": "ONLY IN SRC",
                         "Comments": "",
                         "Common_File": "No",
                         "Src_Size": os.path.getsize(os.path.join(src_dir, file)),
                         "Tgt_Size": "",
                         "Diff_Percentage": ""})

        # Add files only in target directory
        for file in only_in_tgt:
            data.append({"FileName": file,
                         "Status": "ONLY IN TGT",
                         "Comments": "",
                         "Common_File": "No",
                         "Src_Size": "",
                         "Tgt_Size": os.path.getsize(os.path.join(tgt_dir, file)),
                         "Diff_Percentage": ""})

        df = pd.DataFrame(data, columns=["FileName", "Common_File", "Src_Size", "Tgt_Size", "Diff_Percentage", "Status",
                                         "Comments"])

        return df

    @keyword('Compare All Files In Directories With Diffs')
    def files_diff_with_diffs(self, src_dir: str, tgt_dir: str) -> None:
        """
        Compare all files in source and target directories and print the comparison results with diffs for matching files.

        `Args:`
        | *`Name`* | *`Type`* | *`Description`* |
        | `src_dir` | `str` | `The source directory to compare.` |
        | `tgt_dir` | `str` | `TThe target directory to compare.` |

        `Returns:`
        | `None` |  |

        Examples:
        | Files Diff With Diffs | ${source_dir} | ${target_dir} |
        | Files Diff With Diffs | /path/to/source/dir | /path/to/target/dir |
        """

        logger.info(f"Step 0: Starting comparison of files in {src_dir} and {tgt_dir}")

        # Call files_diff to get a dataframe of comparison results
        df = self.files_diff(src_dir, tgt_dir)

        # Loop through the matching files and print diffs
        for _, row in df[df['Status'] == 'Match'].iterrows():
            src_file = os.path.join(src_dir, row['File Name'])
            tgt_file = os.path.join(tgt_dir, row['File Name'])
            self.df_diff(src_file, tgt_file)

    def replace_values_in_string(self, input_string, value_dict):
        """
        Replace values in a string based on a provided dictionary.

        Args:
        | *`Name`* | *`Type`* | *`Description`* |
        | `input_string` | `str` | `The input string to be modified.` |
        | `value_dict` | `dict` | `A dictionary containing keys (values to be replaced) and corresponding values.` |

        Returns:
        | *`Type`* | *`Description`* |
        | `str` | `Modified string.` |
        """
        for key, value in value_dict.items():
            input_string = input_string.replace(f"'~{key}'", f"'{value}'")
        return input_string

    @keyword("Get File Format Using ICD")
    def get_file_format_using_icd(self, icd_in_db: str = 'No', query_file_path: str = None, config_file_path: str = None,
                                  db_config_name: str = None, replace_value_dict: dict = None,
                                  icd_config_path: str = None):
        """
        Get the file format using an ICD configuration.

        Args:
        | Name              | Type | Description                                       |
        |-------------------|------|---------------------------------------------------|
        | icd_in_db         | str  | Indicates whether the ICD configuration is in a database ('yes') or a file. |
        | query_file_path   | str  | The path to the query file if 'icd_in_db' is 'yes'. |
        | config_file_path  | str  | The path to the database configuration file if 'icd_in_db' is 'yes'. |
        | db_config_name    | str  | The name of the database configuration if 'icd_in_db' is 'yes'. |
        | icd_config_path   | str  | The path to the ICD configuration file if 'icd_in_db' is 'no'. |

        Returns:
        | Name           | Type           | Description                                     |
        |----------------|----------------|-------------------------------------------------|
        | colspecs       | list of tuples | List of colspecs as (start_pos, end_pos).        |
        | widths         | list of int    | List of field widths.                           |
        | dtypes         | dict           | Dictionary of field names and corresponding data types. |
        | column_names   | list of str    | List of column names.                           |

        Examples:
        | ${colspecs} | ${widths} | ${data_types} | ${column_names} | = | Get File Format Using ICD | /path/to/icd_config.xlsx |
        """
        colspecs = []
        widths = []
        dtypes = {}
        column_names = []
        key_columns = []

        logger.info(f"ICD in DB: {icd_in_db}")

        if icd_in_db.lower() == 'yes':
            logger.info(f"Step-01: Reading ICD configuration from DB")
            try:
                query = DBUtils.get_query_from_file(query_file_path)
                query = self.replace_values_in_string(query, replace_value_dict)
                icd_df = DBUtils.execute_query(query=query, config_file_path=config_file_path,
                                               database_config_name=db_config_name)
                # Filter rows where unique_key_field is 'Yes'
                key_columns = icd_df[icd_df['unique_key_field'] == 'Y']['field_name'].tolist()
            except Exception as e:
                logger.error(f"Error reading ICD configuration: {e}")
                return None, None, None, None
        else:
            logger.info(f"Step-01: Reading ICD configuration from '{icd_config_path}'")
            try:
                icd_df = pd.read_excel(icd_config_path, sheet_name=0)  # Read the first sheet
                key_columns = icd_df[icd_df['unique_key_field'] == 'Y']['field_name'].tolist()
            except Exception as e:
                logger.error(f"Error reading ICD configuration: {e}")
                return None, None, None, None

            if not all(col in icd_df.columns for col in ['field_name', 'mandatory', 'datatype', 'field_length',
                                                         'start_position', 'end_position']):
                logger.error("ICD configuration is not as per format. Required columns are: "
                             "'field_name', 'mandatory', 'datatype', 'field_length', 'start_position', 'end_position'")
                return None, None, None, None

        for _, row in icd_df.iterrows():
            start_pos = row['start_position'] - 1  # Adjust to 0-index
            end_pos = row['end_position']
            length = row['field_length']
            dtypes = row['datatype']
            field_name = row['field_name']

            colspecs.append((start_pos, end_pos))
            widths.append(length)

            # Map ICD data type to Pandas data type
            # if dtypes.startswith('Varchar'):
            #     dtypes[field_name] = 'str'
            # elif dtypes.startswith('Integer'):
            #     dtypes[field_name] = 'int'
            # elif dtypes.startswith('Float'):
            #     dtypes[field_name] = 'float'

            column_names.append(field_name)

        if any(pd.isna(start) or pd.isna(end) for start, end in colspecs):
            colspecs = None
        dtypes = None

        if icd_in_db.lower() == 'yes':
            logger.info("Step-02: ICD configuration successfully processed")
            return colspecs, widths, dtypes, column_names, key_columns
        else:
            logger.info("Step-02: ICD configuration successfully processed")
            return colspecs, widths, dtypes, column_names, key_columns

    def consolidate_execution_summary(self, results_folder_path, test_data_file_path):
        """
        Consolidate execution summary for tests marked as "tbe" (To Be Executed).

        This function reads test data from a CSV file, filters tests marked as "tbe", and consolidates execution summary
        data from each test. It also extracts test statuses from associated output XML files.

        `Args:`
            | `results_folder_path` | `str` | The path to the folder containing test results. |
            | `test_data_csv` | `str` | The path to the test data CSV file. |

        `Returns:`
            | `pd.DataFrame` | The consolidated DataFrame containing execution summary. |
        """

        try:
            file_extension = os.path.splitext(test_data_file_path)[-1].lower()

            if file_extension == '.csv':
                test_data_df = pd.read_csv(test_data_file_path, dtype='str')
            elif file_extension == '.xlsx':
                test_data_df = pd.read_excel(test_data_file_path, dtype='str', sheet_name=0)
            else:
                raise ValueError("Unsupported file format. Supported formats are .csv and .xlsx.")

            # Filter rows where tbe is "Yes"
            tbe_yes_tests = test_data_df[test_data_df["tbe"] == "Yes"]

            # Initialize an empty DataFrame to store the consolidated data
            consolidated_df = pd.DataFrame()

            step = 1  # Initialize step counter

            for test_name in tbe_yes_tests["test_name"]:
                logger.info(f"Step - {step}: Processing test: {test_name}")
                step += 1

                # Define the path to the Overall Count Statistics.csv file for each test
                summary_file_path = os.path.join(results_folder_path, test_name, "Overall Count Statistics.csv")

                # Check if the file exists
                if os.path.exists(summary_file_path):
                    logger.info(f"Step - {step}: Loading summary data for test: {test_name}")
                    step += 1

                    # Load the data from the CSV file into a DataFrame
                    summary_df = pd.read_csv(summary_file_path)

                    # Add 'Test Name' column to the summary DataFrame
                    summary_df['Test Name'] = test_name

                    # Determine status based on conditions
                    conditions = (
                            (summary_df['Total Keys & Data Matched'] < summary_df['Source Total']) |
                            (summary_df['Total Keys & Data Matched'] < summary_df['Target Total'])
                    )
                    summary_df['Status'] = 'FAIL' if conditions.any() else 'PASS'

                    # Concatenate the data to the consolidated DataFrame
                    consolidated_df = pd.concat([consolidated_df, summary_df], ignore_index=True)
                else:
                    logger.info(f"Step - {step}: Summary file not found for test: {test_name}")
                    step += 1

            # Reorder the columns
            consolidated_df = consolidated_df[
                ['Test Name', 'Status'] + [col for col in consolidated_df if col not in ['Test Name', 'Status']]]

            return consolidated_df
        except Exception as e:
            logger.error(f"Error occurred while reading test data from '{test_data_file_path}': {e}")



import base64
import logging
from typing import Dict, Optional

import pandas as pd
import yaml
import oracledb
from robot.api.deco import keyword
import matplotlib.pyplot as plt
from IPython.display import display
from robot.api import logger
from typing import List

import plotly.offline as pyo
import plotly.graph_objs as go
import plotly.figure_factory as ff
import plotly.subplots as sp

oracledb.init_oracle_client()
pyo.init_notebook_mode()


class DBUtils:

    def __init__(self):
        pass

    @keyword('Get Query from File')
    def get_query_from_file(self, file_path: str) -> str:
        """
        Reads SQL query from file

        Args:
        | *`Name`* | *`Type`* | *`Description`* |
        | `file_path` | `(str)` | `Path to file containing SQL query.` |

        Returns:
        | *`Type`* | *`Description`* |
        | `(str)` | `SQL query as a string.` |
        """
        logger.info('Step 1: Reading SQL query from file...')
        with open(file_path, 'r') as file:
            return file.read()

    def execute_query(self, query: str, config_file_path: str, database_config_name: str,
                      query_params: Optional[Dict] = None) -> pd.DataFrame:
        """
        Executes an SQL query on a specified database configuration using oracledb.

        Args:
        | *`Name`* | *`Type`* | *`Description`* |
        | `query` | `(str)` | `SQL query to execute.` |
        | `config_file_path` | `(str)` | `Path to the configuration file containing the database configurations.` |
        | `database_config_name` | `(str)` | `Name of the database configuration in the config file.` |
        | `query_params` | `(Optional[Dict])` | `Optional dictionary of query parameters.` |

        Returns:
        | *`Type`* | *`Description`* |
        | `pandas.DataFrame` | `A Pandas DataFrame containing the results of the SQL query.` |

        Configuration file structure:
        ```
        database_configurations:
            my_database:
                username: my_username
                password: my_password
                host: my_host
                port: my_port
                service_name: my_service_name
        ```

        Examples:
        | *`Robot Framework`* |                                                                                   |
        | `Execute Query`     | `query=SELECT * FROM my_table` `config_file_path=/path/to/config.yaml` `database_config_name=my_database` |
        |                      | `query_params={"param1": "value1"}`                                        |
        """

        # Step 1: Load database configuration from config file
        logging.info('Step 1: Loading database configuration from config file...')
        with open(config_file_path, 'r') as file:
            config = yaml.safe_load(file)
        db_config = config['database_configurations'][database_config_name]

        # Step 2: Connect to the database using oracledb
        logging.info('Step 2: Connecting to the database using oracledb...')
        connection_params = {
            "user": db_config['username'],
            "password": base64.b64decode(db_config['password'].encode()).decode(),
            "dsn": f"{db_config['host']}:{db_config['port']}/{db_config['service_name']}"
        }
        connection = oracledb.connect(**connection_params)

        # Step 3: Execute SQL query
        logging.info('Step 3: Executing SQL query...')
        with connection.cursor() as cursor:
            if query_params:
                cursor.execute(query, query_params)
            else:
                cursor.execute(query)

            # Fetch all the results into a Pandas DataFrame
            result = pd.DataFrame(cursor.fetchall(), columns=[desc[0] for desc in cursor.description])

        result = result.fillna('')

        # Step 4: Close the connection
        connection.close()

        return result

    @keyword('Visualize Data')
    def visualize_data(self, dataframe: pd.DataFrame, x_col: str, y_col: str, plot_type: str = 'bar',
                       title: str = 'Data Visualization', plot_name: str = 'Trace') -> None:
        """
        Visualizes query results as a plot and displays them as a DataFrame.

        Args:
        | *`Name`* | *`Type`* | *`Description`* |
        | *`dataframe`* | `pandas.DataFrame` | A Pandas DataFrame containing the results of the SQL query. |
        | *`x_col`*     | `str`             | The name of the column to use for the x-axis of the plot. |
        | *`y_col`*     | `str`             | The name of the column to use for the y-axis of the plot. |
        | *`plot_type`*  | `str`             | The type of plot to use. Default is 'bar', but can be 'line', 'scatter', etc. |
        | *`title`*      | `str`             | The title of the plot. Default is 'Data Visualization'. |
        | *`plot_name`*  | `str`             | The name of the trace in the plot. Default is 'Trace'. |

        Examples:
        | *Keyword*              | *Arguments*                                                               |
        | `Visualize Data`       | dataframe=x, x_col=column_name_1, y_col=column_name_2, plot_type='bar', title='Bar Plot', plot_name='Trace' |
        | `Visualize Data`       | dataframe=y, x_col=column_name_3, y_col=column_name_4, plot_type='scatter', title='Scatter Plot', plot_name='Data' |
        | `Visualize Data`       | dataframe=z, x_col=column_name_5, y_col=column_name_6, plot_type='line', title='Line Plot', plot_name='Line Data' |
        """

        # Step 1: Visualize data as plot
        logger.info(f"Step 1: Visualizing data as {plot_type} plot...")

        trace = go.Bar(x=dataframe[x_col], y=dataframe[y_col], name=plot_name)

        data = [trace]
        layout = go.Layout(title=title, barmode='group', height=600, showlegend=True,
                           margin=dict(l=50, r=50, t=50, b=50))
        fig = go.Figure(data=trace, layout=layout)
        fig.update_layout(xaxis_title=x_col.replace('_', ' ').capitalize(),
                          yaxis_title=y_col.replace('_', ' ').capitalize())

        pyo.iplot(fig)

        # Step 2: Create and display the table
        logger.info("Step 2: Displaying data as a DataFrame table...")

        table = ff.create_table(dataframe)
        table.layout.autosize = False
        table.layout.width = 800
        table.layout.align = 'center'  # Set the alignment to center

        pyo.iplot(table)

        # Old code
        # Step 1: Visualize data as plot
        # logger.info(f"Step 1: Visualizing data as {plot_type} plot...")
        # plt.figure(figsize=(15, 5))
        #
        # if plot_type == 'bar':
        #     plt.bar(dataframe[x_col], dataframe[y_col])
        # elif plot_type == 'line':
        #     plt.plot(dataframe[x_col], dataframe[y_col])
        # elif plot_type == 'scatter':
        #     plt.scatter(dataframe[x_col], dataframe[y_col])
        # else:
        #     raise ValueError(f"Invalid plot type '{plot_type}'. Supported types are 'bar', 'line', 'scatter'.")
        #
        # plt.xlabel(x_col)
        # plt.ylabel(y_col)
        # plt.title('Query Results')
        # plt.show()
        #
        # # Step 2: Display data as DataFrame
        # logger.info('Step 2: Displaying data as DataFrame...')
        # display(dataframe)

    @keyword('Execute SQL Query and Visualize Data')
    def execute_sql_query_and_visualize_data(self, query_file_path: str, config_file_path: str,
                                             database_config_name: str,
                                             x_col: str, y_col: str, chart_type: str = 'bar') -> None:
        """
        Executes an SQL query and visualizes the results as a chart.

        Args:
        | *`Name`* | *`Type`* | *`Description`* |
        | *`query_file_path`*     | `str` | Path to the file containing the SQL query to execute.     |
        | *`config_file_path`*    | `str` | Path to the configuration file containing the database configurations. |
        | *`database_config_name`*| `str` | Name of the database configuration in the config file.  |
        | *`x_col`*               | `str` | The name of the column to use for the x-axis of the chart. |
        | *`y_col`*               | `str` | The name of the column to use for the y-axis of the chart. |
        | *`chart_type`*          | `str` | The type of chart to use. Can be one of "bar", "line", "scatter", "histogram". Default is "bar". |

        Examples:
        | *Keyword*                              | *Arguments*                                                       |
        | `Execute SQL Query and Visualize Data` | query_file_path=/path/to/query.sql, config_file_path=/path/to/config.yaml, database_config_name=my_database, x_col=column_name_1, y_col=column_name_2, chart_type=bar |
        """
        # Step 1: Get SQL query from file
        logger.info('Step 1: Get SQL query from file...')
        query = self.get_query_from_file(query_file_path)

        # Step 2: Execute SQL query
        logger.info('Step 2: Execute SQL query...')
        result = self.execute_query(query, config_file_path, database_config_name)

        # Step 3: Visualize data
        logger.info('Step 3: Visualizing data...')
        if chart_type == 'bar':
            self.visualize_data.bar_chart(result, x_col, y_col)
        elif chart_type == 'line':
            self.visualize_data.line_chart(result, x_col, y_col)
        elif chart_type == 'scatter':
            self.visualize_data.scatter_plot(result, x_col, y_col)
        elif chart_type == 'histogram':
            self.visualize_data.histogram(result, x_col, y_col)
        else:
            raise ValueError(
                f'Invalid chart type: {chart_type}. Supported chart types are: "bar", "line", "scatter", "histogram"')

    @keyword('Get DataFrame Columns')
    def get_dataframe_columns(self, dataframe: pd.DataFrame) -> List[str]:
        """
        Returns the column names of a Pandas DataFrame.

        Args:
        | *`Name`* | *`Type`* | *`Description`* |
        | *`dataframe`* | `pandas.DataFrame` | A Pandas DataFrame to get column names from. |

        Returns:
        A list of column names.

        Examples:
        | *Keyword*              | *Arguments*                            |
        | `Get DataFrame Columns`| dataframe=my_dataframe                 |
        """
        # Step 1: Get dataframe columns
        logger.info('Step 1: Getting dataframe columns...')
        return dataframe.columns.tolist()

# *** Settings ***
# Library  MyLibrary.py
#
# *** Variables ***
# ${QUERY_FILE_PATH}  /path/to/query.sql
# ${CONFIG_FILE_PATH}  /path/to/config.yaml
# ${DATABASE_CONFIG_NAME}  my_database
# ${X_COL}  column1
# ${Y_COL}  column2
# ${PLOT_TYPE}  bar
#
# *** Tasks ***
# Get Query from File
#     [Documentation]  Retrieves a query from a SQL file
#     [Task]  ${query}=  Get Query from File  file_path=${QUERY_FILE_PATH}
#     [Return]  ${query}
#
# Execute Query
#     [Documentation]  Executes a SQL query using a database configuration
#     [Task]  ${query}=  Get Query from File  file_path=${QUERY_FILE_PATH}
#     ...  config_file_path=${CONFIG_FILE_PATH}
#     ...  database_config_name=${DATABASE_CONFIG_NAME}
#     ${results}=  Execute Query  query=${query}  config_file_path=${CONFIG_FILE_PATH}  database_config_name=${DATABASE_CONFIG_NAME}
#     [Return]  ${results}
#
# Visualize Data
#     [Documentation]  Generates a visualization of data using a given plot type
#     [Task]  ${query}=  Get Query from File  file_path=${QUERY_FILE_PATH}
#     ...  config_file_path=${CONFIG_FILE_PATH}
#     ...  database_config_name=${DATABASE_CONFIG_NAME}
#     ${results}=  Execute Query  query=${query}  config_file_path=${CONFIG_FILE_PATH}  database_config_name=${DATABASE_CONFIG_NAME}
#     Visualize Data  dataframe=${results}  x_col=${X_COL}  y_col=${Y_COL}  plot_type=${PLOT_TYPE}
#
# Execute Query and Visualize Data
#     [Documentation]  Executes a SQL query using a database configuration and generates a visualization of data
#     [Task]  ${query}=  Get Query from File  file_path=${QUERY_FILE_PATH}
#     ...  config_file_path=${CONFIG_FILE_PATH}
#     ...  database_config_name=${DATABASE_CONFIG_NAME}
#     ${results}=  Execute Query  query=${query}  config_file_path=${CONFIG_FILE_PATH}  database_config_name=${DATABASE_CONFIG_NAME}
#     Visualize Data  dataframe=${results}  x_col=${X_COL}  y_col=${Y_COL}  plot_type=${PLOT_TYPE}



import duckdb
import pandas as pd
import os
import logging
import gc  # Importing garbage collection module
import csv

# Setting up logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

# Use a disk-based database instead of in-memory
DATABASE = 'C:/pythontest/results/my_duckdb_database.db'
# Declare variables
SRC_PATH = 'C:/pythontest/src/5m Sales Records.csv'
TRG_PATH = 'C:/pythontest/trg/5m Sales Records.csv'
MAPPING_FILE = 'IGP_ ICD Mapping.csv'
TABLE_SRC = 'table_src'
TABLE_TRG = 'table_trg'
OUTPUT_EXCEL = 'C:/pythontest/results/comparison_results.xlsx'
PRIMARY_KEYS = ['Order Date']

logging.info('Variables declared')


def read_fixed_length_file(file_path, structure):
    """Read a fixed-length .dat file and return a list of dictionaries."""
    data = []
    with open(file_path, 'r') as file:
        for line in file:
            row = {}
            for start, end, column_name in structure:
                row[column_name] = line[start:end].strip()
            data.append(row)
    return data


logging.info('Function read_fixed_length_file defined')


def load_data_into_duckdb(file_path, table_name):
    """Load data from a file into DuckDB based on its format"""
    extension = os.path.splitext(file_path)[1].lower()
    if extension == '.dat':
        mapping_df = pd.read_csv(MAPPING_FILE)
        mapping_df['IGP Base Field'] = mapping_df['IGP Base Field'].str.strip()
        column_names = mapping_df['IGP Base Field'].tolist()
        structure = list(zip(mapping_df['Start Position'], mapping_df['End Position'], column_names))
        data = read_fixed_length_file(file_path, structure)
        conn.register(table_name, pd.DataFrame(data, columns=column_names))
    elif extension == '.csv':
        conn.execute(f"CREATE TABLE {table_name} AS SELECT * FROM read_csv_auto('{file_path}')")
    elif extension == '.xml':
        data = pd.read_xml(file_path)
        conn.register(table_name, data)
    elif extension == '.json':
        data = pd.read_json(file_path)
        conn.register(table_name, data)
    else:
        raise ValueError(f'Unsupported file format: {extension}')
    column_info = conn.execute(f'PRAGMA table_info({table_name})').fetchall()
    return [col[1] for col in column_info]


logging.info('Function load_data_into_duckdb defined')

# Create a connection to DuckDB
conn = duckdb.connect(database=DATABASE, read_only=False)
logging.info('Connected to DuckDB')

# Load the data into DuckDB based on file format
columns_src_names = load_data_into_duckdb(SRC_PATH, TABLE_SRC)
columns_trg_names = load_data_into_duckdb(TRG_PATH, TABLE_TRG)
logging.info('Data loaded into DuckDB')

# Define keys string for SQL queries
keys_str = ' AND '.join([f'a."{key}" = b."{key}"' if ' ' in key else f'a.{key} = b.{key}' for key in PRIMARY_KEYS])

# List of columns to compare excluding primary keys
compare_columns = [col for col in columns_src_names if col not in PRIMARY_KEYS]

# Compute source_total and target_total
source_total = conn.execute(f"SELECT COUNT(*) FROM {TABLE_SRC}").fetchone()[0]
target_total = conn.execute(f"SELECT COUNT(*) FROM {TABLE_TRG}").fetchone()[0]
logging.info(f'Source total: {source_total}, Target total: {target_total}')


# Data breaks
logging.info('Starting data breaks computation')

total_data_break_query = f"""
SELECT a.*, b.*
FROM {TABLE_SRC} AS a
LEFT JOIN {TABLE_TRG} AS b ON ({keys_str})
"""

total_data_break_data = conn.execute(total_data_break_query).fetchall()

# Determine the indices for the source and target columns
src_indices = [columns_src_names.index(col) for col in compare_columns]
trg_indices = [columns_trg_names.index(col) + len(columns_src_names) for col in compare_columns]

# Data breaks
logging.info('Starting data breaks computation')

total_data_break_query = f"""
SELECT a.*, b.*
FROM {TABLE_SRC} AS a
LEFT JOIN {TABLE_TRG} AS b ON ({keys_str})
"""

# Open a CSV file for writing
with open('total_data_break.csv', 'w', newline='') as csvfile:
    writer = None

    # Fetching results in chunks
    chunk_size = 1000
    result_set = conn.execute(total_data_break_query)
    while True:
        chunk = result_set.fetchmany(chunk_size)
        if not chunk:
            break

        # Process the chunk and prepare data breaks
        total_data_breaks_chunk = []
        for row in chunk:
            mismatched_columns = {}
            for key in PRIMARY_KEYS:
                key_index = columns_src_names.index(key)
                mismatched_columns[key] = row[key_index]

            for idx, col in enumerate(compare_columns):
                src_value = row[src_indices[idx]]
                if trg_indices[idx] < len(row):
                    trg_value = row[trg_indices[idx]]
                else:
                    trg_value = None

                if src_value != trg_value:
                    mismatched_columns[f"{col}_Expected"] = src_value
                    mismatched_columns[f"{col}_Actual"] = trg_value
                    logging.info('src not equal to trg')

            if len(mismatched_columns) > len(PRIMARY_KEYS):
                total_data_breaks_chunk.append(mismatched_columns)

        # Write the chunk to CSV
        if not writer:
            writer = csv.DictWriter(csvfile, fieldnames=total_data_breaks_chunk[0].keys())
            writer.writeheader()
        writer.writerows(total_data_breaks_chunk)

gc.collect()  # Trigger garbage collection
logging.info('Data breaks computation completed')


# Only in Source
only_in_source_query = f"""
SELECT a.*
FROM {TABLE_SRC} AS a
LEFT JOIN {TABLE_TRG} AS b ON ({keys_str})
WHERE b.{PRIMARY_KEYS[0]} IS NULL
"""

only_in_source = conn.execute(only_in_source_query).fetchall()
pd.DataFrame(only_in_source, columns=columns_src_names).to_csv("only_in_source.csv", index=False)

# Only in Target
only_in_target_query = f"""
SELECT b.*
FROM {TABLE_TRG} AS b
LEFT JOIN {TABLE_SRC} AS a ON ({keys_str})
WHERE a.{PRIMARY_KEYS[0]} IS NULL
"""

only_in_target = conn.execute(only_in_target_query).fetchall()
pd.DataFrame(only_in_target, columns=columns_src_names).to_csv("only_in_target.csv", index=False)

logging.info('Computed data for "Only in Source" and "Only in Target"')

# Compute metrics and write to separate CSVs
keys_str_no_alias = ', '.join(PRIMARY_KEYS)

# Source Key Duplicate
source_key_duplicate_query = f"""
SELECT a.*
FROM {TABLE_SRC} AS a
JOIN (
    SELECT {keys_str_no_alias}
    FROM {TABLE_SRC}
    GROUP BY ({keys_str_no_alias})
    HAVING COUNT(*) > 1
) AS b ON a.{PRIMARY_KEYS[0]} = b.{PRIMARY_KEYS[0]}
"""

source_key_duplicate = conn.execute(source_key_duplicate_query).fetchall()
pd.DataFrame(source_key_duplicate, columns=columns_src_names).to_csv('source_key_duplicate.csv', index=False)

# Target Key Duplicate
target_key_duplicate_query = f"""
SELECT a.*
FROM {TABLE_TRG} AS a
JOIN (
    SELECT {keys_str_no_alias}
    FROM {TABLE_TRG}
    GROUP BY ({keys_str_no_alias})
    HAVING COUNT(*) > 1
) AS b ON a.{PRIMARY_KEYS[0]} = b.{PRIMARY_KEYS[0]}
"""

target_key_duplicate = conn.execute(target_key_duplicate_query).fetchall()
pd.DataFrame(target_key_duplicate, columns=columns_trg_names).to_csv('target_key_duplicate.csv', index=False)

logging.info('Computed data for "Source Key Duplicate" and "Target Key Duplicate"')

# Create summary dictionary
summary_data = {
    'Source Total': source_total,
    'Target Total': target_total,
    'Columns with Differences': len(column_differences),
    'Only in Source': len(only_in_source),
    'Only in Target': len(only_in_target),
    'Source Key Duplicate': len(source_key_duplicate),
    'Target Key Duplicate': len(target_key_duplicate)
}

pd.DataFrame([summary_data]).to_csv('summary.csv', index=False)
logging.info('Summary data computed and saved')

# Close the DuckDB connection
conn.close()
logging.info('DuckDB connection closed')
